"""
Test basic markdown renderings
"""


from tests.utils import (assert_render, render_markdown, row_text,
                         setup_lookatme)


def test_headings(tmpdir, mocker):
    """Test basic header rendering
    """
    setup_lookatme(tmpdir, mocker, style={
        "headings": {
            "default": {
                "fg": "bold",
                "bg": "",
                "prefix": "|",
                "suffix": "|",
            },
        },
    })

    rendered = render_markdown("""
# H1
## H2
### H3
---
""")

    stripped_rows = [
        b"",
        b"|H1|",
        b"",
        b"|H2|",
        b"",
        b"|H3|",
        b"",
    ]
    assert_render(stripped_rows, rendered)


def test_table(tmpdir, mocker):
    """Test basic table rendering
    """
    setup_lookatme(tmpdir, mocker, style={
        "table": {
            "column_spacing": 1,
            "header_divider": "-",
        },
    })

    rendered = render_markdown("""
| H1      |   H2   |     H3 |
|:--------|:------:|-------:|
| 1value1 | value2 | value3 |
| 1       | 2      | 3      |
""")

    stripped_rows = [
        b"H1        H2       H3",
        b"------- ------ ------",
        b"1value1 value2 value3",
        b"1          2        3",
    ]
    assert_render(stripped_rows, rendered, full_strip=True)


def test_lists(tmpdir, mocker):
    """Test list rendering
    """
    setup_lookatme(tmpdir, mocker, style={
        "bullets": {
            "default": "*",
            "1": "-",
            "2": "=",
            "3": "^",
        },
    })

    rendered = render_markdown("""
* list 1
  * list 2
    * list 3
      * list 4
  * list 2
    * list 3
    * list 3

* list 2
""")

    stripped_rows = [
        b'',
        b'  - list 1',
        b'    = list 2',
        b'      ^ list 3',
        b'        * list 4',
        b'    = list 2',
        b'      ^ list 3',
        b'      ^ list 3',
        b'  - list 2',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_lists_with_newline(tmpdir, mocker):
    """Test list rendering with a newline between a new nested list and the
    previous list item
    """
    setup_lookatme(tmpdir, mocker, style={
        "bullets": {
            "default": "*",
            "1": "-",
            "2": "=",
            "3": "^",
        },
    })

    rendered = render_markdown("""
* list 1

  * list 2
""")

    stripped_rows = [
        b'',
        b'  - list 1',
        b'',
        b'    = list 2',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_numbered_lists(tmpdir, mocker):
    """Test list rendering
    """
    setup_lookatme(tmpdir, mocker, style={
        "bullets": {
            "default": "*",
            "1": "-",
            "2": "=",
            "3": "^",
        },
        "numbering": {
            "default": "numeric",
            "1": "numeric",
            "2": "alpha",
            "3": "roman",
        },
    })

    rendered = render_markdown("""
1. list 1
    1. alpha1
    1. alpha2
    1. alpha3
1. list 2
    1. alpha1.1
        1. roman1
        1. roman2
        1. roman3
    1. alpha1.2
        * test1
        * test2
1. list 3
""")

    stripped_rows = [
        b'',
        b'  1. list 1',
        b'     a. alpha1',
        b'     b. alpha2',
        b'     c. alpha3',
        b'  2. list 2',
        b'     a. alpha1.1',
        b'        i.   roman1',
        b'        ii.  roman2',
        b'        iii. roman3',
        b'     b. alpha1.2',
        b'        ^ test1',
        b'        ^ test2',
        b'  3. list 3',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_hrule(tmpdir, mocker):
    """Test that hrules render correctly
    """
    setup_lookatme(tmpdir, mocker, style={
        "hrule": {
            "style": {
                "fg": "",
                "bg": "",
            },
            "char": "=",
        },
    })

    rendered = render_markdown("---", width=10, single_slide=True)
    stripped_rows = [
        b'',
        b'==========',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_block_quote(tmpdir, mocker):
    """Test block quote rendering
    """
    setup_lookatme(tmpdir, mocker, style={
        "quote": {
            "style": {
                "fg": "",
                "bg": "",
            },
            "side": ">",
            "top_corner": "-",
            "bottom_corner": "=",
        },
    })

    rendered = render_markdown("""
> this is a quote
""")

    stripped_rows = [
        b'',
        b'-',
        b'>  this is a quote',
        b'=',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_code(tmpdir, mocker):
    """Test code block rendering
    """
    setup_lookatme(tmpdir, mocker, style={
        "style": "monokai",
    })

    rendered = render_markdown("""
```python
def some_fn(*args, **kargs):
    pass```
""")

    stripped_rows = [
        b'',
        b'def some_fn(*args, **kargs):',
        b'    pass',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_empty_codeblock(tmpdir, mocker):
    """Test that empty code blocks render correctly
    """
    setup_lookatme(tmpdir, mocker, style={
        "style": "monokai",
    })

    render_markdown("""
```python

```""")


def test_code_yaml(tmpdir, mocker):
    """Test code block rendering with yaml language
    """
    setup_lookatme(tmpdir, mocker, style={
        "style": "monokai",
    })

    rendered = render_markdown("""
```yaml
test: a value
test2: "another value"
array:
  - item1
  - item2
  - item3
```""")

    stripped_rows = [
        b'',
        b'test: a value',
        b'test2: "another value"',
        b'array:',
        b'  - item1',
        b'  - item2',
        b'  - item3',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_inline(tmpdir, mocker):
    """Test inline markdown
    """
    setup_lookatme(tmpdir, mocker, style={
        "style": "monokai",
        "link": {
            "fg": "underline",
            "bg": "default",
        },
    })

    rendered = render_markdown("*emphasis*")
    assert rendered[1][0][0].foreground == "default,italics"
    assert row_text(rendered[1]).strip() == b"emphasis"

    rendered = render_markdown("**emphasis**")
    assert rendered[1][0][0].foreground == "default,underline"
    assert row_text(rendered[1]).strip() == b"emphasis"

    rendered = render_markdown("_emphasis_")
    assert rendered[1][0][0].foreground == "default,italics"
    assert row_text(rendered[1]).strip() == b"emphasis"

    rendered = render_markdown("__emphasis__")
    assert rendered[1][0][0].foreground == "default,underline"
    assert row_text(rendered[1]).strip() == b"emphasis"

    rendered = render_markdown("`inline code`")
    assert row_text(rendered[1]).rstrip() == b" inline code"

    rendered = render_markdown("~~strikethrough~~")
    assert rendered[1][0][0].foreground == "default,strikethrough"
    assert row_text(rendered[1]).rstrip() == b"strikethrough"

    rendered = render_markdown("[link](http://domain.tld)")
    assert rendered[1][0][0].foreground == "default,underline"
    assert row_text(rendered[1]).rstrip() == b"link"

    rendered = render_markdown("http://domain.tld")
    assert rendered[1][0][0].foreground == "default,underline"
    assert row_text(rendered[1]).rstrip() == b"http://domain.tld"

    rendered = render_markdown("![link](http://domain.tld)")
    assert rendered[1][0][0].foreground == "default,underline"
    assert row_text(rendered[1]).rstrip() == b"link"



"""
This module handles loading and using lookatme_contrib modules

    def loads(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).loads(*args, **kwargs)
        if res is None:
            raise ValueError("Could not loads")
        return res

    def load(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).load(*args, **kwargs)
        if res is None:
            raise ValueError("Could not load")
        return res

Contrib modules are directly used
"""


import functools
from typing import List

import lookatme.ascii_art
import lookatme.prompt
from lookatme.exceptions import IgnoredByContrib

CONTRIB_MODULES = []


def validate_extension_mod(_ext_name, ext_mod):
    """Validate the extension, returns an array of warnings associated with the
    module
    """
    res = []
    if not hasattr(ext_mod, "user_warnings"):
        res.append("'user_warnings' is missing. Extension is not able to "
                   "provide user warnings.")
    else:
        res += ext_mod.user_warnings()

    return res


def load_contribs(contrib_names, safe_contribs, ignore_load_failure=False):
    """Load all contrib modules specified by ``contrib_names``. These should
    all be namespaced packages under the ``lookatmecontrib`` namespace. E.g.
    ``lookatmecontrib.calendar`` would be an extension provided by a
    contrib module, and would be added to an ``extensions`` list in a slide's
    YAML header as ``calendar``.

    ``safe_contribs`` is a set of contrib names that are manually provided
    by the user by the ``-e`` flag or env variable of extensions to auto-load.
    """
    if contrib_names is None:
        return

    errors = []
    all_warnings = []
    for contrib_name in contrib_names:
        module_name = f"lookatme.contrib.{contrib_name}"
        try:
            mod = __import__(module_name, fromlist=[contrib_name])
        except Exception as e:
            if ignore_load_failure:
                continue
            errors.append(str(e))
        else:
            if contrib_name not in safe_contribs:
                ext_warnings = validate_extension_mod(contrib_name, mod)
                if len(ext_warnings) > 0:
                    all_warnings.append((contrib_name, ext_warnings))
            CONTRIB_MODULES.append(mod)

    _handle_load_errors_warnings(errors, all_warnings)


def _handle_load_errors_warnings(errors: List[str], all_warnings: List[str]):
    """Handle all load errors and warnings from loading contrib modules
    """
    if len(errors) > 0:
        raise Exception(
            "Error loading one or more extensions:\n\n" + "\n".join(errors),
        )

    if len(all_warnings) == 0:
        return

    print("\nExtension-provided user warnings:")
    for ext_name, ext_warnings in all_warnings:
        print("\n  {!r}:\n".format(ext_name))
        for ext_warning in ext_warnings:
            print("    * {}".format(ext_warning))
    print("")

    if not lookatme.prompt.yes("Continue anyways?"):
        exit(1)


def contrib_first(fn):
    """A decorator that allows contrib modules to override default behavior
    of lookatme. E.g., a contrib module may override how a table is displayed
    to enable sorting, or enable displaying images rendered with ANSII color
    codes and box drawing characters, etc.

    Contrib modules may ignore chances to override default behavior by raising
    the ``lookatme.contrib.IgnoredByContrib`` exception.
    """
    fn_name = fn.__name__

    @functools.wraps(fn)
    def inner(*args, **kwargs):
        for mod in CONTRIB_MODULES:
            if not hasattr(mod, fn_name):
                continue
            try:
                return getattr(mod, fn_name)(*args, **kwargs)
            except IgnoredByContrib:
                pass

        return fn(*args, **kwargs)

    return inner


def shutdown_contribs():
    """Call the shutdown function on all contrib modules
    """
    for mod in CONTRIB_MODULES:
        getattr(mod, "shutdown", lambda: 1)()



"""
"""


import urwid


def row_text(rendered_row):
    """Return all text joined together from the rendered row
    """
    return b"".join(x[-1] for x in rendered_row)


def resolve_bag_of_text_markup_or_widgets(items):
    """Resolve the list of items into either contiguous urwid.Text() instances,
    or pre-existing urwid.Widget objects
    """
    res = []
    curr_text_markup = []
    for item in items:
        if isinstance(item, tuple) or isinstance(item, str):
            curr_text_markup.append(item)
        else:
            if len(curr_text_markup) > 0:
                res.append(urwid.Text(curr_text_markup))
                curr_text_markup = []
            res.append(item)

    if len(curr_text_markup) > 0:
        res.append(urwid.Text(curr_text_markup))

    return res


def dict_deep_update(to_update, new_vals):
    """Deeply update the to_update dict with the new_vals
    """
    for key, value in new_vals.items():
        if isinstance(value, dict):
            node = to_update.setdefault(key, {})
            dict_deep_update(node, value)
        else:
            to_update[key] = value


def spec_from_style(styles):
    """Create an urwid.AttrSpec from a {fg:"", bg:""} style dict. If styles
    is a string, it will be used as the foreground
    """
    if isinstance(styles, str):
        return urwid.AttrSpec(styles, "")
    else:
        return urwid.AttrSpec(styles.get("fg", ""), styles.get("bg", ""))


def get_fg_bg_styles(style):
    if style is None:
        return [], []

    def non_empty_split(data):
        res = [x.strip() for x in data.split(",")]
        return list(filter(None, res))

    # from lookatme.config.get_style()
    if isinstance(style, dict):
        return non_empty_split(style["fg"]), non_empty_split(style["bg"])
    # just a str will only set the foreground color
    elif isinstance(style, str):
        return non_empty_split(style), []
    elif isinstance(style, urwid.AttrSpec):
        return non_empty_split(style.foreground), non_empty_split(style.background)
    else:
        raise ValueError("Unsupported style value {!r}".format(style))


def overwrite_spec(orig_spec, new_spec):
    if orig_spec is None:
        orig_spec = urwid.AttrSpec("", "")
    if new_spec is None:
        new_spec = urwid.AttrSpec("", "")

    fg_orig = orig_spec.foreground.split(",")
    fg_orig_color = orig_spec._foreground_color()
    fg_orig.remove(fg_orig_color)

    bg_orig = orig_spec.background.split(",")
    bg_orig_color = orig_spec._background()
    bg_orig.remove(bg_orig_color)

    fg_new = new_spec.foreground.split(",")
    fg_new_color = new_spec._foreground_color()
    fg_new.remove(fg_new_color)

    bg_new = new_spec.background.split(",")
    bg_new_color = new_spec._background()
    bg_new.remove(bg_new_color)

    if fg_new_color == "default":
        fg_orig.append(fg_orig_color)
    else:
        fg_new.append(fg_new_color)

    if bg_new_color == "default":
        bg_orig.append(bg_orig_color)
    else:
        bg_new.append(bg_new_color)

    return urwid.AttrSpec(
        ",".join(set(fg_orig + fg_new)),
        ",".join(set(bg_orig + bg_new)),
    )


def flatten_text(text, new_spec=None):
    """Return a flattend list of tuples that can be used as the first argument
    to a new urwid.Text().

    :param urwid.Text text: The text to flatten
    :param urwid.AttrSpec new_spec: A new spec to merge with existing styles
    :returns: list of tuples
    """
    text, chunk_stylings = text.get_text()

    res = []
    total_len = 0
    for spec, chunk_len in chunk_stylings:
        split_text = text[total_len:total_len + chunk_len]
        total_len += chunk_len

        split_text_spec = overwrite_spec(new_spec, spec)
        res.append((split_text_spec, split_text))

    if len(text[total_len:]) > 0:
        res.append((new_spec, text[total_len:]))

    return res


def can_style_item(item):
    """Return true/false if ``style_text`` can work with the given item
    """
    return isinstance(item, (urwid.Text, list, tuple))


def styled_text(text, new_styles, old_styles=None, supplement_style=False):
    """Return a styled text tuple that can be used within urwid.Text.

    .. note::

        If an urwid.Text instance is passed in as the ``text`` parameter,
        alignment values will be lost and must be explicitly re-added by the
        caller.
    """
    if isinstance(text, urwid.Text):
        new_spec = spec_from_style(new_styles)
        return flatten_text(text, new_spec)
    elif (isinstance(text, tuple)
            and isinstance(text[0], urwid.AttrSpec)
            and isinstance(text[1], urwid.Text)):
        text = text[1].text
        old_styles = text[0]

    new_fg, new_bg = get_fg_bg_styles(new_styles)
    old_fg, old_bg = get_fg_bg_styles(old_styles)

    def join(items):
        return ",".join(set(items))

    spec = urwid.AttrSpec(
        join(new_fg + old_fg),
        join(new_bg + old_bg),
    )
    return (spec, text)


def pile_or_listbox_add(container, widgets):
    """Add the widget/widgets to the container
    """
    if isinstance(container, urwid.ListBox):
        return listbox_add(container, widgets)
    elif isinstance(container, urwid.Pile):
        return pile_add(container, widgets)
    else:
        raise ValueError("Container was not listbox, nor pile")


def listbox_add(listbox, widgets):
    if not isinstance(widgets, list):
        widgets = [widgets]

    for w in widgets:
        if len(listbox.body) > 0 \
                and isinstance(w, urwid.Divider) \
                and isinstance(listbox.body[-1], urwid.Divider):
            continue
        listbox.body.append(w)


def pile_add(pile, widgets):
    """
    """
    if not isinstance(widgets, list):
        widgets = [widgets]

    for w in widgets:
        if len(pile.contents) > 0 \
                and isinstance(w, urwid.Divider) \
                and isinstance(pile.contents[-1][0], urwid.Divider):
            continue
        pile.contents.append((w, pile.options()))


def int_to_roman(integer):
    integer = int(integer)
    ints = [1000, 900,  500, 400, 100,  90, 50,  40, 10,  9,   5,  4,   1]
    nums = ["m",  "cm", "d", "cd", "c", "xc",
            "l", "xl", "x", "ix", "v", "iv", "i"]
    result = []
    for i in range(len(ints)):
        count = integer // ints[i]
        result.append(nums[i] * count)
        integer -= ints[i] * count
    return "".join(result)



"""
Defines utilities for testing lookatme
"""


import urwid

import lookatme.config
import lookatme.tui
from lookatme.parser import Parser


def setup_lookatme(tmpdir, mocker, style=None):
    mocker.patch.object(lookatme.config, "LOG")
    mocker.patch("lookatme.config.SLIDE_SOURCE_DIR", new=str(tmpdir))

    if style is not None:
        mocker.patch("lookatme.config.STYLE", new=style)


def assert_render(correct_render, rendered, full_strip=False):
    for idx, row in enumerate(rendered):
        if full_strip:
            stripped = row_text(row).strip()
        else:
            stripped = row_text(row).rstrip()
        if idx >= len(correct_render):
            assert stripped == b""
        else:
            assert correct_render[idx] == stripped


def render_markdown(markdown, height=50, width=200, single_slide=False):
    """Returns the rendered canvas contents of the markdown
    """
    loop = urwid.MainLoop(urwid.ListBox([]))
    renderer = lookatme.tui.SlideRenderer(loop)
    renderer.start()

    parser = Parser(single_slide=single_slide)
    _, slides = parser.parse_slides({"title": ""}, markdown)

    renderer.stop()
    contents = renderer.render_slide(slides[0], force=True)
    renderer.join()

    container = urwid.ListBox([urwid.Text("testing")])
    container.body = contents
    return list(container.render((width, height)).content())


def spec_and_text(item):
    """``item`` should be an item from a rendered widget, a tuple of the form

    .. code-block:: python

        (spec, ?, text)
    """
    return item[0], item[2]


def row_text(rendered_row):
    """Return all text joined together from the rendered row
    """
    return b"".join(x[-1] for x in rendered_row)



"""
Test tha lookatme YAML schemas behave as expected
"""


import datetime

import marshmallow.exceptions
import pytest
from marshmallow import Schema, fields

import lookatme.schemas as schemas


def test_meta_schema():
    """Test the meta schema
    """
    title = "TITLE"
    author = "AUTHOR"
    date = "2019-01-01"
    yaml_text = f"""
title: {title}
author: {author}
date: {date}
    """

    schema = schemas.MetaSchema().loads(yaml_text)
    assert schema["title"] == title
    assert schema["author"] == author
    assert schema["date"] == "2019-01-01"


def test_meta_schema_allowed_extra_top_level():
    """Test that extra top-level fields are allowed in the metadata. A separate
    test will test that the style metadata is strictly validated.
    """
    """Test the meta schema
    """
    title = "TITLE"
    author = "AUTHOR"
    date = "2019-01-01"
    yaml_text = f"""
title: {title}
author: {author}
date: {date}
tags: [t1, t2, t3]
status: status_str
    """

    schema = schemas.MetaSchema().loads(yaml_text)
    assert schema["title"] == title
    assert schema["author"] == author
    assert schema["date"] == "2019-01-01"
    assert schema["tags"] == ["t1", "t2", "t3"]
    assert schema["status"] == "status_str"


def test_meta_schema_strict_style_validation():
    """Test that the style schema is STRICTLY validated. Not having this will
    make it hard to debug why mistakes in style names aren't having the desired
    effect on lookatme. We want the user to know what fields they got wrong
    in the style.
    """
    """Test the meta schema
    """
    title = "TITLE"
    author = "AUTHOR"
    date = "2019-01-01"
    yaml_text = f"""
title: {title}
author: {author}
date: {date}
styles:
  invalid: 100
    """
    with pytest.raises(marshmallow.exceptions.ValidationError) as exc_info:
        schemas.MetaSchema().loads(yaml_text)

    assert exc_info.value.messages_dict \
        == {"styles": {"invalid": ["Unknown field."]}}


def _validate_field_recursive(path, field, gend_value):
    """Only validate the leaf nodes - we want *specific* values that have
    changed!
    """
    if isinstance(field, Schema):
        for field_name, sub_field in field.fields.items():
            _validate_field_recursive(
                f"{path}.{field_name}", sub_field, gend_value[field_name])
    elif isinstance(field, fields.Nested):
        if field.dump_default is None:
            nested_field = field.nested()  # type: ignore
            _validate_field_recursive(path, nested_field, gend_value)
        else:
            for field_name, sub_field in field.dump_default.items():
                _validate_field_recursive(
                    f"{path}.{field_name}", sub_field, gend_value[field_name])
    elif isinstance(field, fields.Field):
        if isinstance(field.dump_default, datetime.datetime):
            return
        assert field.dump_default == gend_value, f"Default value not correct at {path}"
    elif isinstance(field, dict):
        for field_name, sub_field in field.items():
            _validate_field_recursive(
                f"{path}.{field_name}", sub_field, gend_value[field_name])
    else:
        assert field == gend_value, f"Default value not correct at {path}"


def test_sanity_check_that_errors_are_detected():
    """Perform a sanity check that we can actually catch errors in generating
    the default schema values.
    """
    schema = schemas.MetaSchema()

    # force a discrepancy in the
    gend_default = schemas.MetaSchema().dump(None)
    gend_default["styles"]["padding"]["left"] = 100

    with pytest.raises(AssertionError) as excinfo:
        _validate_field_recursive(
            "__root__.styles", schema.fields["styles"], gend_default["styles"])
    assert "Default value not correct at __root__.styles.padding" in str(
        excinfo)


def test_styles_defaults():
    """Ensure that style value defaults are generated correctly
    """
    schema = schemas.MetaSchema()
    gend_default = schemas.MetaSchema().dump(None)
    _validate_field_recursive(
        "__root__.styles", schema.fields["styles"], gend_default["styles"])


def test_meta_defaults():
    """Test that the default values in the schema are actually used
    """
    schema = schemas.MetaSchema()
    gend_default = schemas.MetaSchema().dump(None)
    _validate_field_recursive("__root__", schema, gend_default)



"""
Defines the built-in styles for lookatme
"""


from lookatme.schemas import StyleSchema
from lookatme.utils import dict_deep_update


def ensure_defaults(mod):
    """Ensure that all required attributes exist within the provided module
    """
    defaults = StyleSchema().dump(None)
    dict_deep_update(defaults, mod.theme)
    return defaults



"""
This module defines a built-in contrib module that enables terminal embedding
within a slide.
"""


import re
import shlex
from typing import Dict

import urwid
import yaml
from marshmallow import Schema, fields

import lookatme.config
import lookatme.render.markdown_block
from lookatme.exceptions import IgnoredByContrib


def user_warnings():
    """Provide warnings to the user that loading this extension may cause
    shell commands specified in the markdown to be run.
    """
    return [
        "Code-blocks with a language starting with 'terminal' will cause shell",
        "  commands from the source markdown to be run",
        "See https://lookatme.readthedocs.io/en/latest/builtin_extensions/terminal.html",
        "  for more details",
    ]


class YamlRender:
    @staticmethod
    def loads(data): return yaml.safe_load(data)
    @staticmethod
    def dumps(data): return yaml.safe_dump(data)


class TerminalExSchema(Schema):
    """The schema used for ``terminal-ex`` code blocks.
    """
    command = fields.Str()
    rows = fields.Int(dump_default=10, load_default=10)
    init_text = fields.Str(dump_default=None, load_default=None)
    init_wait = fields.Str(dump_default=None, load_default=None)
    init_codeblock = fields.Bool(dump_default=True, load_default=True)
    init_codeblock_lang = fields.Str(dump_default="text", load_default="text")

    class Meta:
        render_module = YamlRender

    def loads(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).loads(*args, **kwargs)
        if res is None:
            raise ValueError("Could not loads")
        return res

    def load(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).load(*args, **kwargs)
        if res is None:
            raise ValueError("Could not load")
        return res


CREATED_TERMS = []


def render_code(token, body, stack, loop):
    lang = token["lang"] or ""

    numbered_term_match = re.match(r'terminal(\d+)', lang)
    if lang != "terminal-ex" and numbered_term_match is None:
        raise IgnoredByContrib

    if numbered_term_match is not None:
        term_data = TerminalExSchema().load({
            "command": token["text"].strip(),
            "rows": int(numbered_term_match.group(1)),
            "init_codeblock": False,
        })

    else:
        term_data = TerminalExSchema().loads(token["text"])

        if term_data["init_text"] is not None and term_data["init_wait"] is not None:
            term_data["command"] = " ".join([shlex.quote(x) for x in [
                "expect", "-c", ";".join([
                    'spawn -noecho {}'.format(term_data["command"]),
                    'expect {{{}}}'.format(term_data["init_wait"]),
                    'send {{{}}}'.format(term_data["init_text"]),
                    'interact',
                    'exit',
                ])
            ]])

    term = urwid.Terminal(
        shlex.split(term_data["command"].strip()),
        main_loop=loop,
        encoding="utf8",
    )
    CREATED_TERMS.append(term)

    line_box = urwid.LineBox(urwid.BoxAdapter(term, height=term_data["rows"]))
    line_box.no_cache = ["render"]

    res = []

    if term_data["init_codeblock"] is True:
        fake_token = {
            "text": term_data["init_text"],
            "lang": term_data["init_codeblock_lang"],
        }
        res += lookatme.render.markdown_block.render_code(
            fake_token, body, stack, loop
        )

    res += [
        urwid.Divider(),
        line_box,
        urwid.Divider(),
    ]

    return res


def shutdown():
    for idx, term in enumerate(CREATED_TERMS):
        lookatme.config.get_log().debug(
            f"Terminating terminal {idx+1}/{len(CREATED_TERMS)}")
        if term.pid is not None:
            term.terminate()



"""
This module tests the Table widget in lookatme/widgets/table.py
"""


import pytest

import lookatme.widgets.table
import tests.utils as utils

TEST_STYLE = {
    "style": "monokai",
    "table": {
        "column_spacing": 3,
        "header_divider": "&",
    },
}


@pytest.fixture(autouse=True)
def table_setup(tmpdir, mocker):
    utils.setup_lookatme(tmpdir, mocker, style=TEST_STYLE)


def test_basic_render(tmpdir, mocker):
    """Test that a Table widget renders correctly
    """
    headers = ["H1", "H2", "H3"]
    aligns = ["left", "center", "right"]
    rows = [
        ["1", "22", "333"],
        ["*1*", "~~22~~", "**333**"],
    ]

    table = lookatme.widgets.table.Table(rows, headers=headers, aligns=aligns)
    canvas = table.render((20,))
    content = list(canvas.content())

    # four rows:
    #  1 headers
    #  2 DIVIDER
    #  3 row1
    #  4 row2
    assert len(content) == 4

    header_row = content[0]
    spec, text = utils.spec_and_text(header_row[0])
    assert "bold" in spec.foreground
    assert text == b"H1"
    spec, text = utils.spec_and_text(header_row[2])
    assert "bold" in spec.foreground
    assert text == b"H2"
    spec, text = utils.spec_and_text(header_row[5])
    assert "bold" in spec.foreground
    assert text == b"H3"

    divider_row = content[1]
    spec, text = utils.spec_and_text(divider_row[0])
    # no styling applied to the divider
    assert spec is None
    assert text == b"&&"
    spec, text = utils.spec_and_text(divider_row[2])
    # no styling applied to the divider
    assert spec is None
    assert text == b"&&"
    spec, text = utils.spec_and_text(divider_row[4])
    # no styling applied to the divider
    assert spec is None
    assert text == b"&&&"

    content_row1 = content[2]
    spec, text = utils.spec_and_text(content_row1[0])
    # no styling applied to this row
    assert spec is None
    assert text == b"1 "
    spec, text = utils.spec_and_text(content_row1[2])
    # no styling applied to this row
    assert spec is None
    assert text == b"22"
    spec, text = utils.spec_and_text(content_row1[4])
    # no styling applied to this row
    assert spec is None
    assert text == b"333"

    content_row1 = content[3]
    spec, text = utils.spec_and_text(content_row1[0])
    # no styling applied to this row
    assert "italics" in spec.foreground
    assert text == b"1"
    spec, text = utils.spec_and_text(content_row1[3])
    # no styling applied to this row
    assert "strikethrough" in spec.foreground
    assert text == b"22"
    spec, text = utils.spec_and_text(content_row1[5])
    # no styling applied to this row
    assert "underline" in spec.foreground
    assert text == b"333"


def test_table_no_headers(mocker):
    """This situation could never happen as parsed from Markdown. See
    https://stackoverflow.com/a/17543474.

    However this situation could happen manually when using the Table() class
    directly.
    """
    headers = None
    aligns = ["left", "center", "right"]
    rows = [
        ["1", "22", "333"],
        ["*1*", "~~22~~", "**333**"],
    ]

    table = lookatme.widgets.table.Table(rows, headers=headers, aligns=aligns)
    canvas = table.render((20,))
    content = list(canvas.content())

    assert len(content) == 2


def test_ignored_extra_column(mocker):
    """Test that extra columns beyond header values are ignored
    """
    headers = ["H1", "H2", "H3"]
    aligns = ["left", "center", "right"]
    rows = [
        ["1", "2", "3"],
        ["1", "2", "3", "4"],
        ["1", "2", "3", "4", "5"],
    ]

    table = lookatme.widgets.table.Table(rows, headers=headers, aligns=aligns)
    canvas = table.render((20,))
    content = list(canvas.content())

    # number of rows of output
    assert len(content) == 5
    assert b"4" not in utils.row_text(content[-2])

    assert b"4" not in utils.row_text(content[-1])
    assert b"5" not in utils.row_text(content[-1])



"""
This module tests contrib-specific functionality
"""

import urwid
from six.moves import StringIO, reload_module

import lookatme.config
import lookatme.contrib
import lookatme.contrib.file_loader
import lookatme.contrib.terminal
import lookatme.pres
import lookatme.render.markdown_block
import lookatme.render.markdown_inline
import lookatme.tui


def setup_contrib(fake_mod):
    lookatme.contrib.CONTRIB_MODULES = [
        lookatme.contrib.terminal,
        lookatme.contrib.file_loader,
        fake_mod
    ]
    reload_module(lookatme.render.markdown_block)
    reload_module(lookatme.render.markdown_inline)
    reload_module(lookatme.tui)


def test_overridable_root(mocker):
    """Ensure that the root urwid component is overridable
    """
    lookatme.config.LOG = mocker.Mock()

    class Wrapper(urwid.WidgetWrap):
        pass

    class FakeMod:
        @staticmethod
        def root_urwid_widget(to_wrap):
            return Wrapper(to_wrap)

    setup_contrib(FakeMod)
    input_stream = StringIO("test")
    pres = lookatme.pres.Presentation(input_stream, "dark")
    tui = lookatme.tui.MarkdownTui(pres)

    assert isinstance(tui.loop.widget, Wrapper)



"""
This module defines a built-in contrib module that enables external files to
be included within the slide. This is extremely useful when having source
code displayed in a code block, and then running/doing something with the
source data in a terminal on the same slide.
"""


import os
import subprocess
from typing import Dict

import yaml
from marshmallow import Schema, fields

import lookatme.config
from lookatme.exceptions import IgnoredByContrib


def user_warnings():
    """Provide warnings to the user that loading this extension may cause
    shell commands specified in the markdown to be run.
    """
    return [
        "Code-blocks with a language starting with 'file' may cause shell",
        "  commands from the source markdown to be run if the 'transform'",
        "  field is set",
        "See https://lookatme.readthedocs.io/en/latest/builtin_extensions/file_loader.html",
        "  for more details",
    ]


class YamlRender:
    @staticmethod
    def loads(data): return yaml.safe_load(data)
    @staticmethod
    def dumps(data): return yaml.safe_dump(data)


class LineRange(Schema):
    start = fields.Integer(dump_default=0, load_default=0)
    end = fields.Integer(dump_default=None, load_default=None)


class FileSchema(Schema):
    path = fields.Str()
    relative = fields.Boolean(dump_default=True, load_default=True)
    lang = fields.Str(dump_default="auto", load_default="auto")
    transform = fields.Str(dump_default=None, load_default=None)
    lines = fields.Nested(
        LineRange,
        dump_default=LineRange().dump(None),
        load_default=LineRange().dump(None)
    )

    class Meta:
        render_module = YamlRender

    def loads(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).loads(*args, **kwargs)
        if res is None:
            raise ValueError("Could not loads")
        return res

    def load(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).load(*args, **kwargs)
        if res is None:
            raise ValueError("Could not load")
        return res


def transform_data(transform_shell_cmd, input_data):
    """Transform the ``input_data`` using the ``transform_shell_cmd``
    shell command.
    """
    proc = subprocess.Popen(
        transform_shell_cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
    )
    stdout, _ = proc.communicate(input=input_data)
    return stdout


def render_code(token, body, stack, loop):
    """Render the code, ignoring all code blocks except ones with the language
    set to ``file``.
    """
    lang = token["lang"] or ""
    if lang != "file":
        raise IgnoredByContrib

    file_info_data = token["text"]
    file_info = FileSchema().loads(file_info_data)

    # relative to the slide source
    if file_info["relative"]:
        base_dir = lookatme.config.SLIDE_SOURCE_DIR
    else:
        base_dir = os.getcwd()

    full_path = os.path.join(base_dir, file_info["path"])
    if not os.path.exists(full_path):
        token["text"] = "File not found"
        token["lang"] = "text"
        raise IgnoredByContrib

    with open(full_path, "rb") as f:
        file_data = f.read()

    if file_info["transform"] is not None:
        file_data = transform_data(file_info["transform"], file_data)

    lines = file_data.split(b"\n")
    lines = lines[file_info["lines"]["start"]:file_info["lines"]["end"]]
    file_data = b"\n".join(lines)
    token["text"] = file_data
    token["lang"] = file_info["lang"]
    raise IgnoredByContrib



#!/usr/bin/env python3

"""
This is the main CLI for lookatme
"""


import io
import os
import tempfile

import click
import pygments.styles

import lookatme
import lookatme.config
import lookatme.log
import lookatme.tui
from lookatme.pres import Presentation
from lookatme.schemas import StyleSchema


@click.command("lookatme")
@click.option("--debug", "debug", is_flag="True", default=False)
@click.option(
    "-l",
    "--log",
    "log_path",
    type=click.Path(writable=True),
    default=os.path.join(tempfile.gettempdir(), "lookatme.log"),
)
@click.option(
    "-t",
    "--theme",
    "theme",
    type=click.Choice(["dark", "light"]),
    default="dark",
)
@click.option(
    "--style",
    "code_style",
    default=None,
    type=click.Choice(list(pygments.styles.get_all_styles())),
)
@click.option(
    "--dump-styles",
    help="Dump the resolved styles that will be used with the presentation to stdout",
    is_flag=True,
    default=False,
)
@click.option(
    "--live",
    "--live-reload",
    "live_reload",
    help="Watch the input filename for modifications and automatically reload",
    is_flag=True,
    default=False,
)
@click.option(
    "-s",
    "--safe",
    help="Do not load any new extensions specified in the source markdown. "
         "Extensions specified via env var or -e are still loaded",
    is_flag=True,
    default=False,
)
@click.option(
    "--no-ext-warn",
    help="Load new extensions specified in the source markdown without warning",
    is_flag=True,
    default=False,
)
@click.option(
    "-i",
    "--ignore-ext-failure",
    help="Ignore load failures of extensions",
    is_flag=True,
    default=False,
)
@click.option(
    "-e",
    "--exts",
    "extensions",
    help="A comma-separated list of extension names to automatically load"
         " (LOOKATME_EXTS)",
    envvar="LOOKATME_EXTS",
    default="",
)
@click.option(
    "--single",
    "--one",
    "single_slide",
    help="Render the source as a single slide",
    is_flag=True,
    default=False
)
@click.version_option(lookatme.__version__)
@click.argument(
    "input_files",
    type=click.File("r"),
    nargs=-1,
)
def main(debug, log_path, theme, code_style, dump_styles,
         input_files, live_reload, extensions, single_slide, safe, no_ext_warn,
         ignore_ext_failure):
    """lookatme - An interactive, terminal-based markdown presentation tool.

    See https://lookatme.readthedocs.io/en/v2.4.1 for documentation
    """
    if debug:
        lookatme.config.LOG = lookatme.log.create_log(log_path)
    else:
        lookatme.config.LOG = lookatme.log.create_null_log()

    if len(input_files) == 0:
        input_files = [io.StringIO("")]

    preload_exts = [x.strip() for x in extensions.split(',')]
    preload_exts = list(filter(lambda x: x != '', preload_exts))
    pres = Presentation(
        input_files[0],
        theme,
        code_style,
        live_reload=live_reload,
        single_slide=single_slide,
        preload_extensions=preload_exts,
        safe=safe,
        no_ext_warn=no_ext_warn,
        ignore_ext_failure=ignore_ext_failure,
    )

    if dump_styles:
        print(StyleSchema().dumps(pres.styles))
        return 0

    try:
        pres.run()
    except Exception as e:
        number = pres.get_tui().curr_slide.number + 1
        click.echo(f"Error rendering slide {number}: {e}")
        if not debug:
            click.echo("Rerun with --debug to view the full traceback in logs")
        else:
            lookatme.config.get_log().exception(
                f"Error rendering slide {number}: {e}")
            click.echo(f"See {log_path} for traceback")
        raise click.Abort()


if __name__ == "__main__":
    main()



"""
Config module for lookatme
"""


import logging
import os
from typing import Dict, Optional

LOG = None
STYLE: Optional[Dict] = None


def get_log() -> logging.Logger:
    if LOG is None:
        raise Exception("LOG was None")
    return LOG


def get_style() -> Dict:
    if STYLE is None:
        raise Exception("STYLE was None!")
    return STYLE


# default to the current working directory - this will be set later by
# pres:Presentation when reading the input stream
SLIDE_SOURCE_DIR = os.getcwd()



"""
Defines a basic Table widget for urwid
"""


from collections import defaultdict
from typing import List, Optional

import urwid

import lookatme.config as config
from lookatme.render.markdown_block import render_text
from lookatme.utils import styled_text
from lookatme.widgets.clickable_text import ClickableText


class Table(urwid.Pile):
    """Create a table from a list of headers, alignment values, and rows.
    """

    signals = ["change"]

    def __init__(self, rows, headers=None, aligns: Optional[List[str]] = None):
        """Create a new table

        :param list columns: The rows to use for the table
        :param list headers: (optional) Headers for the table
        :param list aligns: (optional) Alignment values for each column
        """
        self.table_rows = rows
        self.table_headers = headers

        if headers is not None:
            self.num_columns = len(headers)
        elif headers is None:
            self.num_columns = len(rows[0])
        else:
            raise ValueError(
                "Invalid table specification: could not determine # of columns"
            )

        if aligns is None:
            aligns = ["left"] * self.num_columns
        self.table_aligns = aligns

        def header_modifier(cell):
            return ClickableText(styled_text(cell.text, "bold"), align=cell.align)

        if self.table_headers is not None:
            self.rend_headers = self.create_cells(
                [self.table_headers], modifier=header_modifier
            )
        else:
            self.rend_headers = []
        self.rend_rows = self.create_cells(self.table_rows)

        self.column_maxes = self.calc_column_maxes()

        cell_spacing = config.get_style()["table"]["column_spacing"]
        self.total_width = sum(self.column_maxes.values()) + (
            cell_spacing * (self.num_columns - 1)
        )

        # final rows
        final_rows = []

        # put headers in Columns
        if self.table_headers is not None:
            header_columns = []
            for idx, header in enumerate(self.rend_headers[0]):
                header = header[0]
                header_with_div = urwid.Pile([
                    self.watch(header),
                    urwid.Divider(config.get_style()[
                                  "table"]["header_divider"]),
                ])
                header_columns.append(
                    (self.column_maxes[idx], header_with_div))
            final_rows.append(urwid.Columns(header_columns, cell_spacing))

        for rend_row in self.rend_rows:
            row_columns = []
            for cell_idx, rend_cell in enumerate(rend_row):
                rend_widgets = [self.watch(rend_widget)
                                for rend_widget in rend_cell]
                rend_pile = urwid.Pile(rend_widgets)
                row_columns.append((self.column_maxes[cell_idx], rend_pile))

            column_row = urwid.Columns(row_columns, cell_spacing)
            final_rows.append(column_row)

        urwid.Pile.__init__(self, final_rows)

    def render(self, *args, **kwargs):
        """Do whatever needs to be done to render the table
        """
        self.set_column_maxes()
        return urwid.Pile.render(self, *args, **kwargs)

    def watch(self, w):
        """Watch the provided widget w for changes
        """
        if "change" not in getattr(w, "signals", []):
            return w

        def wrapper(*_, **__):
            self._invalidate()
            self._emit("change")

        urwid.connect_signal(w, "change", wrapper)
        return w

    def _invalidate(self):
        self.set_column_maxes()
        urwid.Pile._invalidate(self)

    def set_column_maxes(self):
        """Calculate and set the column maxes for this table
        """
        self.column_maxes = self.calc_column_maxes()
        cell_spacing = config.get_style()["table"]["column_spacing"]
        self.total_width = sum(self.column_maxes.values()) + (
            cell_spacing * (self.num_columns - 1)
        )

        for columns, info in self.contents:
            # row should be a Columns instance
            new_columns = []
            for idx, column_items in enumerate(columns.contents):
                column_widget, column_info = column_items
                new_columns.append((
                    column_widget,
                    (column_info[0], self.column_maxes[idx], column_info[2]),
                ))
            columns.contents = new_columns

    def calc_column_maxes(self):
        column_maxes = defaultdict(int)
        for row in self.rend_headers + self.rend_rows:
            for idx, cell in enumerate(row):
                for widget in cell:
                    if not isinstance(widget, urwid.Text):
                        widg_len = 15
                    else:
                        widg_len = len(widget.text)
                    if idx > self.num_columns:
                        break
                    column_maxes[idx] = max(column_maxes[idx], widg_len)
        return column_maxes

    def create_cells(self, body_rows, modifier=None):
        """Create the rows for the body, optionally calling a modifier function
        on each created cell Text. The modifier must accept an urwid.Text object
        and must return an urwid.Text object.
        """
        res = []

        for row in body_rows:
            rend_row = []
            for idx, cell in enumerate(row):
                if idx >= self.num_columns:
                    break
                rend_cell_widgets = render_text(text=cell)
                new_widgets = []
                for widget in rend_cell_widgets:
                    if isinstance(widget, urwid.Text):
                        widget.align = self.table_aligns[idx] or "left"
                        if modifier is not None:
                            widget = modifier(widget)
                    new_widgets.append(widget)
                rend_row.append(new_widgets)
            res.append(rend_row)

        return res



"""
Logging module
"""


import logging


def create_log(log_path):
    """Create a new log that writes to log_path
    """
    logging.basicConfig(filename=log_path, level=logging.DEBUG)
    return logging.getLogger("lookatme")


def create_null_log():
    """Create a logging object that does nothing
    """
    logging.basicConfig(handlers=[logging.NullHandler()])
    return logging.getLogger("lookatme")



"""
Basic user-prompting helper functions
"""


def yes(msg):
    """Prompt the user for a yes/no answer. Returns bool
    """
    answer = input("{} (Y/N) ".format(msg))
    return answer.strip().lower() in ["y", "yes"]



"""
This module tests that markdown presentations can be correctly parsed
"""


import datetime

from lookatme.parser import Parser


def test_parse_metadata():
    """Test that metadata can be correctly parsed out of a markdown
    presentation
    """
    title = "Presentation Title"
    date = "September 2, 2022"
    author = "The Author"

    input_data = f"""
---
title: {title}
date: {date}
author: {author}
---
remaining
    """

    parser = Parser()
    input_data, meta = parser.parse_meta(input_data)
    assert input_data.strip() == "remaining"
    assert meta["title"] == title
    assert meta["date"] == "September 2, 2022"
    assert meta["author"] == author


def test_parse_metadata_empty():
    """Test that metadata can be correctly parsed out of a markdown
    presentation
    """
    input_data = """
---
---
remaining
    """

    parser = Parser()
    input_data, meta = parser.parse_meta(input_data)
    assert input_data.strip() == "remaining"
    now = datetime.datetime.now()
    assert meta["title"] == ""
    assert meta["date"] == now.strftime("%Y-%m-%d")
    assert meta["author"] == ""


def test_parse_slides():
    """Test that slide parsing works correctly
    """
    input_data = r"""
# Slide 1

* list
  * item
  * item
  * item

Hello there this is a paragraph

```python
code block
```

---

# Slide 2

More text
    """
    parser = Parser()
    slides = parser.parse_slides({}, input_data)
    assert len(slides) == 2


def test_parse_smart_slides_one_h1():
    """Test that slide smart splitting works correctly
    """
    input_data = r"""
# Heading Title

## Heading 2

some text

## Heading 3

more text

## Heading 4

### Sub heading

#### Sub Heading
    """
    parser = Parser()
    meta = {"title": ""}
    _, slides = parser.parse_slides(meta, input_data)
    assert len(slides) == 4
    assert meta["title"] == "Heading Title"


def test_parse_smart_slides_multiple_h2():
    """Test that slide smart splitting works correctly
    """
    input_data = r"""
## Heading 2

some text

## Heading 3

more text

## Heading 4

### Sub heading

#### Sub Heading
    """
    parser = Parser()
    meta = {"title": ""}
    _, slides = parser.parse_slides(meta, input_data)
    assert len(slides) == 3
    assert meta["title"] == ""



"""
This module defines the parser for the markdown presentation file
"""


import re
from collections import defaultdict
from typing import AnyStr, Callable, Dict, List, Tuple

import mistune

from lookatme.schemas import MetaSchema
from lookatme.slide import Slide


def is_progressive_slide_delimiter_token(token):
    """Returns True if the token indicates the end of a progressive slide

    :param dict token: The markdown token
    :returns: True if the token is a progressive slide delimiter
    """
    return token["type"] == "close_html" and re.match(r'<!--\s*stop\s*-->', token["text"])


class Parser(object):
    """A parser for markdown presentation files
    """

    def __init__(self, single_slide=False):
        """Create a new Parser instance
        """
        self._single_slide = single_slide

    def parse(self, input_data):
        """Parse the provided input data into a Presentation object

        :param str input_data: The input markdown presentation to parse
        :returns: Presentation
        """
        input_data, meta = self.parse_meta(input_data)
        input_data, slides = self.parse_slides(meta, input_data)
        return meta, slides

    def parse_slides(self, meta, input_data):
        """Parse the Slide out of the input data

        :param dict meta: The parsed meta values
        :param str input_data: The input data string
        :returns: tuple of (remaining_data, slide)
        """
        # slides are delimited by ---
        md = mistune.Markdown()

        state = {}
        tokens = md.block.parse(input_data, state)

        num_hrules, hinfo = self._scan_for_smart_split(tokens)
        keep_split_token = True

        if self._single_slide:
            def slide_split_check(_):  # type: ignore
                return False

            def heading_mod(_):  # type: ignore
                pass
        elif num_hrules == 0:
            if meta.get("title", "") in ["", None]:
                meta["title"] = hinfo["title"]

            def slide_split_check(token):
                nonlocal hinfo
                return (
                    token["type"] == "heading"
                    and token["level"] == hinfo["lowest_non_title"]
                )

            def heading_mod(token):
                nonlocal hinfo
                token["level"] = max(
                    token["level"] - (hinfo["title_level"] or 0),
                    1,
                )
            keep_split_token = True
        else:
            def slide_split_check(token):
                return token["type"] == "hrule"

            def heading_mod(token):
                pass
            keep_split_token = False

        slides = self._split_tokens_into_slides(
            tokens, slide_split_check, heading_mod, keep_split_token)

        return "", slides

    def _split_tokens_into_slides(
            self,
            tokens: List[Dict],
            slide_split_check: Callable,
            heading_mod: Callable,
            keep_split_token: bool
    ) -> List[Slide]:
        """Split the provided tokens into slides using the slide_split_check
        and heading_mod arguments.
        """
        slides = []
        curr_slide_tokens = []
        for token in tokens:
            should_split = slide_split_check(token)
            if token["type"] == "heading":
                heading_mod(token)

            # new slide!
            if should_split:
                if keep_split_token and len(slides) == 0 and len(curr_slide_tokens) == 0:
                    pass
                else:
                    slides.extend(self._create_slides(
                        curr_slide_tokens, len(slides)))
                curr_slide_tokens = []
                if keep_split_token:
                    curr_slide_tokens.append(token)
                continue
            else:
                curr_slide_tokens.append(token)

        slides.extend(self._create_slides(curr_slide_tokens, len(slides)))

        return slides

    def _scan_for_smart_split(self, tokens):
        """Scan the provided tokens for the number of hrules, and the lowest
        (h1 < h2) header level.

        :returns: tuple (num_hrules, lowest_header_level)
        """
        hinfo = {
            "title_level": None,
            "lowest_non_title": 10,
            "counts": defaultdict(int),
            "title": "",
        }
        num_hrules = 0
        first_heading = None
        for token in tokens:
            if token["type"] == "hrule":
                num_hrules += 1
            elif token["type"] == "heading":
                hinfo["counts"][token["level"]] += 1
                if first_heading is None:
                    first_heading = token

        # started off with the lowest heading, make this title
        if (
            hinfo["counts"]
            and first_heading
            and hinfo["counts"][first_heading["level"]] == 1
        ):
            hinfo["title"] = first_heading["text"]
            del hinfo["counts"][first_heading["level"]]
            hinfo["title_level"] = first_heading["level"]

        low_level = min(list(hinfo["counts"].keys()) + [10])
        hinfo["title_level"] = low_level - 1
        hinfo["lowest_non_title"] = low_level

        return num_hrules, hinfo

    def parse_meta(self, input_data) -> Tuple[AnyStr, Dict]:
        """Parse the PresentationMeta out of the input data

        :param str input_data: The input data string
        :returns: tuple of (remaining_data, meta)
        """
        found_first = False
        yaml_data = []
        skipped_chars = 0
        for line in input_data.split("\n"):
            skipped_chars += len(line) + 1
            stripped_line = line.strip()

            is_marker = (re.match(r'----*', stripped_line) is not None)
            if is_marker:
                if not found_first:
                    found_first = True
                # found the second one
                else:
                    break

            if found_first and not is_marker:
                yaml_data.append(line)
                continue

            # there was no ----* marker
            if not found_first and stripped_line != "":
                break

        if not found_first:
            return input_data, MetaSchema().load_partial_styles({}, partial=True)

        new_input = input_data[skipped_chars:]
        if len(yaml_data) == 0:
            return new_input, MetaSchema().load_partial_styles({}, partial=True)

        yaml_data = "\n".join(yaml_data)
        data = MetaSchema().loads_partial_styles(yaml_data, partial=True)
        return new_input, data

    def _create_slides(self, tokens, number):
        """Iterate on tokens and create slides out of them. Can create multiple
        slides if the tokens contain progressive slide delimiters.

        :param list tokens: The tokens to create slides out of
        :param int number: The starting slide number
        :returns: A list of Slides
        """
        slide_tokens = []
        for token in tokens:
            if is_progressive_slide_delimiter_token(token):
                yield Slide(slide_tokens[:], number)
                number += 1
            else:
                slide_tokens.append(token)
        yield Slide(slide_tokens, number)



"""
Test the main CLI
"""


from typing import Optional

import yaml
from click.testing import CliRunner

import lookatme
import lookatme.schemas
import lookatme.themes.dark as dark_theme
import lookatme.themes.light as light_theme
import lookatme.tui
from lookatme.__main__ import main


def run_cmd(*args):
    """Run the provided arguments
    """
    runner = CliRunner()
    return runner.invoke(main, args)


def test_dump_styles_unicode():
    """Test that dump styles works correctly
    """
    res = run_cmd("--dump-styles")
    assert res.exit_code == 0
    assert "█" in res.output


def _get_dumped_style(
    tmpdir,
    theme: Optional[str] = None,
    md_meta_style: Optional[str] = None,
    cli_style: Optional[str] = None
) -> str:
    cli_args = ["--dump-styles"]

    if theme is not None:
        cli_args += ["--theme", theme]
    if md_meta_style is not None:
        tmpfile = tmpdir.join("test.md")
        with open(tmpfile, "w") as f:
            f.write("\n".join([
                "---",
                "styles:",
                "  style: {}".format(md_meta_style),
                "---",
            ]))
        cli_args += [str(tmpfile)]
    if cli_style is not None:
        cli_args += ["--style", cli_style]

    res = run_cmd(*cli_args)
    assert res.exit_code == 0

    yaml_data = yaml.safe_load(res.output)
    return yaml_data["style"]


def test_style_override_precedence_dark(tmpdir):
    """Test that dump styles works correctly
    """
    default_style = _get_dumped_style(tmpdir)
    themed_style = _get_dumped_style(tmpdir, theme="dark")
    themed_and_md = _get_dumped_style(
        tmpdir,
        theme="dark",
        md_meta_style="emacs"
    )
    themed_and_md_and_cli = _get_dumped_style(
        tmpdir,
        theme="dark",
        md_meta_style="emacs",
        cli_style="zenburn"
    )

    default = lookatme.schemas.MetaSchema().dump(None)
    assert default_style == default["styles"]["style"]

    dark_theme_styles = lookatme.schemas.StyleSchema().dump(dark_theme.theme)
    assert themed_style == dark_theme_styles["style"]  # type: ignore

    assert themed_and_md == "emacs"
    assert themed_and_md_and_cli == "zenburn"


def test_style_override_precedence_light(tmpdir):
    """Test that dump styles works correctly
    """
    default_style = _get_dumped_style(tmpdir)
    themed_style = _get_dumped_style(tmpdir, theme="light")
    themed_and_md = _get_dumped_style(
        tmpdir,
        theme="light",
        md_meta_style="emacs"
    )
    themed_and_md_and_cli = _get_dumped_style(
        tmpdir,
        theme="light",
        md_meta_style="emacs",
        cli_style="zenburn"
    )

    default = lookatme.schemas.MetaSchema().dump(None)
    assert default_style == default["styles"]["style"]

    light_theme_styles = lookatme.schemas.StyleSchema().dump(light_theme.theme)
    assert themed_style == light_theme_styles["style"]  # type: ignore

    assert themed_and_md == "emacs"
    assert themed_and_md_and_cli == "zenburn"


def test_version():
    """Test the version option
    """
    res = run_cmd("--version")
    assert res.exit_code == 0
    assert lookatme.__version__ in res.output


def test_exceptions(tmpdir, mocker):
    """Test exception handling on invalid inputs
    """
    log_path = tmpdir.join("log.txt")
    pres_path = tmpdir.join("test.md")
    with pres_path.open("w") as f:
        f.write("# Hello")

    slide_number = 3
    exception_text = "EXCEPTION TEXT"

    def fake_create_tui(*args, **kwargs):
        res = mocker.MagicMock()
        res.curr_slide.number = slide_number
        res.run.side_effect = Exception(exception_text)
        return res
    mocker.patch.object(lookatme.tui, "create_tui", fake_create_tui)

    res = run_cmd("--log", str(log_path), str(pres_path))
    assert exception_text in res.output
    assert f"slide {slide_number+1}" in res.output
    # should remind us to rerun with --debug to see the traceback
    assert "--debug" in res.output

    res = run_cmd("--debug", "--log", str(log_path), str(pres_path))
    assert exception_text in res.output
    assert f"slide {slide_number+1}" in res.output
    # should remind us to check log_path for the traceback
    assert str(log_path) in res.output



"""
Defines render functions that render lexed markdown block tokens into urwid
representations
"""


import mistune
import urwid

import lookatme.config as config
import lookatme.render.markdown_inline as markdown_inline_renderer
import lookatme.render.pygments as pygments_render
import lookatme.utils as utils
from lookatme.contrib import contrib_first
from lookatme.widgets.clickable_text import ClickableText


def _meta(item):
    if not hasattr(item, "meta"):
        meta = {}
        setattr(item, "meta", meta)
    else:
        meta = getattr(item, "meta")
    return meta


def _set_is_list(item, level=1, ordered=False):
    _meta(item).update({
        "is_list": True,
        "list_level": level,
        "ordered": ordered,
        "item_count": 0,
    })


def _inc_item_count(item):
    _meta(item)["item_count"] += 1
    return _meta(item)["item_count"]


def _is_list(item):
    return _meta(item).get("is_list", False)


def _list_level(item):
    return _meta(item).get("list_level", 1)


@contrib_first
def render_newline(token, body, stack, loop):
    """Render a newline

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    return urwid.Divider()


@contrib_first
def render_hrule(token, body, stack, loop):
    """Render a newline

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    hrule_conf = config.get_style()["hrule"]
    div = urwid.Divider(hrule_conf['char'], top=1, bottom=1)
    return urwid.Pile([urwid.AttrMap(div, utils.spec_from_style(hrule_conf['style']))])


@contrib_first
def render_heading(token, body, stack, loop):
    """Render markdown headings, using the defined styles for the styling and
    prefix/suffix.

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.

    Below are the default stylings for headings:

    .. code-block:: yaml

        headings:
          '1':
            bg: default
            fg: '#9fc,bold'
            prefix: "██ "
            suffix: ""
          '2':
            bg: default
            fg: '#1cc,bold'
            prefix: "▓▓▓ "
            suffix: ""
          '3':
            bg: default
            fg: '#29c,bold'
            prefix: "▒▒▒▒ "
            suffix: ""
          '4':
            bg: default
            fg: '#66a,bold'
            prefix: "░░░░░ "
            suffix: ""
          default:
            bg: default
            fg: '#579,bold'
            prefix: "░░░░░ "
            suffix: ""

    :returns: A list of urwid Widgets or a single urwid Widget
    """
    headings = config.get_style()["headings"]
    level = token["level"]
    style = config.get_style()["headings"].get(str(level), headings["default"])

    prefix = utils.styled_text(style["prefix"], style)
    suffix = utils.styled_text(style["suffix"], style)

    rendered = render_text(text=token["text"])
    if len(rendered) > 0:
        rendered = rendered[0]

    return [
        urwid.Divider(),
        ClickableText(
            [prefix] + utils.styled_text(rendered, style) + [suffix]),  # type: ignore
        urwid.Divider(),
    ]


@contrib_first
def render_table(token, body, stack, loop):
    """Renders a table using the :any:`Table` widget.

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.

    The table widget makes use of the styles below:

    .. code-block:: yaml

        table:
          column_spacing: 3
          header_divider: "─"

    :returns: A list of urwid Widgets or a single urwid Widget
    """
    from lookatme.widgets.table import Table

    headers = token["header"]
    aligns = token["align"]
    cells = token["cells"]

    table = Table(cells, headers=headers, aligns=aligns)
    padding = urwid.Padding(table, width=table.total_width + 2, align="center")

    def table_changed(*args, **kwargs):
        padding.width = table.total_width + 2

    urwid.connect_signal(table, "change", table_changed)

    return padding


@contrib_first
def render_list_start(token, body, stack, loop):
    """Handles the indentation when starting rendering a new list. List items
    themselves (with the bullets) are rendered by the
    :any:`render_list_item_start` function.

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    res = urwid.Pile([])

    in_list = _is_list(stack[-1])
    list_level = 1
    if in_list:
        list_level = _list_level(stack[-1]) + 1
    _set_is_list(res, list_level, ordered=token['ordered'])
    _meta(res)['list_start_token'] = token
    _meta(res)['max_list_marker_width'] = token.get('max_list_marker_width', 2)
    stack.append(res)

    widgets = []
    if not in_list:
        widgets.append(urwid.Divider())
        widgets.append(urwid.Padding(res, left=2))
        widgets.append(urwid.Divider())
        return widgets
    return res


@contrib_first
def render_list_end(token, body, stack, loop):
    """Pops the pushed ``urwid.Pile()`` from the stack (decreases indentation)

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    meta = _meta(stack[-1])
    meta['list_start_token']['max_list_marker_width'] = meta['max_list_marker_width']
    stack.pop()


def _list_item_start(token, body, stack, loop):
    """Render the start of a list item. This function makes use of two
    different styles, one each for unordered lists (bullet styles) and ordered
    lists (numbering styles):

    .. code-block:: yaml

        bullets:
          '1': "•"
          '2': "⁃"
          '3': "◦"
          default: "•"
        numbering:
          '1': "numeric"
          '2': "alpha"
          '3': "roman"
          default: "numeric"

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    list_level = _list_level(stack[-1])
    curr_count = _inc_item_count(stack[-1])
    pile = urwid.Pile(urwid.SimpleFocusListWalker([]))

    meta = _meta(stack[-1])

    if meta["ordered"]:
        numbering = config.get_style()["numbering"]
        list_marker_type = numbering.get(str(list_level), numbering["default"])
        sequence = {
            "numeric": lambda x: str(x),
            "alpha": lambda x: chr(ord("a") + x - 1),
            "roman": lambda x: utils.int_to_roman(x),
        }[list_marker_type]
        list_marker = sequence(curr_count) + "."
    else:
        bullets = config.get_style()["bullets"]
        list_marker = bullets.get(str(list_level), bullets["default"])

    marker_text = list_marker + " "
    if len(marker_text) > meta["max_list_marker_width"]:
        meta["max_list_marker_width"] = len(marker_text)
    marker_col_width = meta['max_list_marker_width']

    res = urwid.Text(("bold", marker_text))
    res = urwid.Columns([
        (marker_col_width, urwid.Text(("bold", marker_text))),
        pile,
    ])
    stack.append(pile)
    return res


@contrib_first
def render_list_item_start(token, body, stack, loop):
    """Render the start of a list item. This function makes use of the styles:

    .. code-block:: yaml

        bullets:
          '1': "•"
          '2': "⁃"
          '3': "◦"
          default: "•"

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    return _list_item_start(token, body, stack, loop)


@contrib_first
def render_loose_item_start(token, body, stack, loop):
    """Render the start of a list item. This function makes use of the styles:

    .. code-block:: yaml

        bullets:
          '1': "•"
          '2': "⁃"
          '3': "◦"
          default: "•"

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    return _list_item_start(token, body, stack, loop)


@contrib_first
def render_list_item_end(token, body, stack, loop):
    """Pops the pushed ``urwid.Pile()`` from the stack (decreases indentation)

    See :any:`lookatme.tui.SlideRenderer.do_render` for argument and return
    value descriptions.
    """
    stack.pop()


@contrib_first
def render_text(token=None, body=None, stack=None, loop=None, text=None):
    """Renders raw text. This function uses the inline markdown lexer
    from mistune with the :py:mod:`lookatme.render.markdown_inline` render module
    to render the lexed inline markup to a list composed of widgets or
    `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_.
    The created list of widgets/Text markup is then used to create and return
    a list composed entirely of widgets and :any:`ClickableText` instances.

    Many other functions call this function directly, passing in the extra
    ``text`` argument and leaving all other arguments blank.

    See :any:`lookatme.tui.SlideRenderer.do_render` for additional argument and
    return value descriptions.
    """
    if text is None and token is not None:
        text = token["text"]

    inline_lexer = mistune.InlineLexer(markdown_inline_renderer)
    res = inline_lexer.output(text)
    if len(res) == 0:
        res = [""]

    widget_list = []
    curr_text_spec = []
    for item in res:
        if isinstance(item, urwid.Widget):
            if len(curr_text_spec) > 0:
                widget_list.append(ClickableText(curr_text_spec))
                curr_text_spec = []
            widget_list.append(item)
        else:
            curr_text_spec.append(item)
    if len(curr_text_spec) > 0:
        widget_list.append(ClickableText(curr_text_spec))

    return widget_list


@contrib_first
def render_paragraph(token, body, stack, loop):
    """Renders the provided text with additional pre and post paddings.

    See :any:`lookatme.tui.SlideRenderer.do_render` for additional argument and
    return value descriptions.
    """
    token["text"] = token["text"].replace("\r\n", " ").replace("\n", " ")
    res = render_text(token, body, stack, loop)
    return [urwid.Divider()] + res + [urwid.Divider()]


@contrib_first
def render_block_quote_start(token, body, stack, loop):
    """Begins rendering of a block quote. Pushes a new ``urwid.Pile()`` to the
    stack that is indented, has styling applied, and has the quote markers
    on the left.

    This function makes use of the styles:

    .. code-block:: yaml

        quote:
          top_corner: "┌"
          bottom_corner: "└"
          side: "╎"
          style:
            bg: default
            fg: italics,#aaa

    See :any:`lookatme.tui.SlideRenderer.do_render` for additional argument and
    return value descriptions.
    """
    pile = urwid.Pile([])
    stack.append(pile)

    styles = config.get_style()["quote"]

    quote_side = styles["side"]
    quote_top_corner = styles["top_corner"]
    quote_bottom_corner = styles["bottom_corner"]
    quote_style = styles["style"]

    return [
        urwid.Divider(),
        urwid.LineBox(
            urwid.AttrMap(
                urwid.Padding(pile, left=2),
                utils.spec_from_style(quote_style),
            ),
            lline=quote_side, rline="",
            tline=" ", trcorner="", tlcorner=quote_top_corner,
            bline=" ", brcorner="", blcorner=quote_bottom_corner,
        ),
        urwid.Divider(),
    ]


@contrib_first
def render_block_quote_end(token, body, stack, loop):
    """Pops the block quote start ``urwid.Pile()`` from the stack, taking
    future renderings out of the block quote styling.

    See :any:`lookatme.tui.SlideRenderer.do_render` for additional argument and
    return value descriptions.
    """
    pile = stack.pop()

    # remove leading/trailing divider if they were added to the pile
    if isinstance(pile.contents[0][0], urwid.Divider):
        pile.contents = pile.contents[1:]
    if isinstance(pile.contents[-1][0], urwid.Divider):
        pile.contents = pile.contents[:-1]


@contrib_first
def render_code(token, body, stack, loop):
    """Renders a code block using the Pygments library.

    See :any:`lookatme.tui.SlideRenderer.do_render` for additional argument and
    return value descriptions.
    """
    lang = token.get("lang", "text") or "text"
    text = token["text"]
    res = pygments_render.render_text(text, lang=lang)

    return [
        urwid.Divider(),
        res,
        urwid.Divider()
    ]



"""Pygments related rendering
"""


import time

import pygments
import pygments.lexers
import pygments.styles
import pygments.util
import urwid
from pygments.formatter import Formatter

import lookatme.config as config

LEXER_CACHE = {}
STYLE_CACHE = {}
FORMATTER_CACHE = {}


def get_formatter(style_name):
    style = get_style(style_name)

    formatter, style_bg = FORMATTER_CACHE.get(style_name, (None, None))
    if formatter is None:
        style_bg = UrwidFormatter.findclosest(
            style.background_color.replace("#", ""))
        formatter = UrwidFormatter(
            style=style,
            usebg=(style_bg is not None),
        )
        FORMATTER_CACHE[style_name] = (formatter, style_bg)
    return formatter, style_bg


def get_lexer(lang, default="text"):
    lexer = LEXER_CACHE.get(lang, None)
    if lexer is None:
        try:
            lexer = pygments.lexers.get_lexer_by_name(lang)
        except pygments.util.ClassNotFound:
            lexer = pygments.lexers.get_lexer_by_name(default)
        LEXER_CACHE[lang] = lexer
    return lexer


def get_style(style_name):
    style = STYLE_CACHE.get(style_name, None)
    if style is None:
        style = pygments.styles.get_style_by_name(style_name)
        STYLE_CACHE[style_name] = style
    return style


def render_text(text, lang="text", style_name=None, plain=False):
    """Render the provided text with the pygments renderer
    """
    if style_name is None:
        style_name = config.get_style()["style"]

    lexer = get_lexer(lang)
    formatter, style_bg = get_formatter(style_name)

    start = time.time()
    code_tokens = lexer.get_tokens(text)
    config.get_log().debug(
        f"Took {time.time()-start}s to render {len(text)} bytes")

    markup = []
    for x in formatter.formatgenerator(code_tokens):
        if style_bg:
            x[0].background = style_bg
        markup.append(x)

    if markup[-1][1] == "\n":
        markup = markup[:-1]

    if len(markup) == 0:
        markup = [(None, "")]
    elif markup[-1][1].endswith("\n"):
        markup[-1] = (markup[-1][0], markup[-1][1][:-1])

    if plain:
        return markup
    else:
        return urwid.AttrMap(urwid.Text(markup), urwid.AttrSpec("default", style_bg))


class UrwidFormatter(Formatter):
    """Formatter that returns [(text,attrspec), ...],
    where text is a piece of text, and attrspec is an urwid.AttrSpec"""

    def __init__(self, **options):
        """Extra arguments:

        usebold: if false, bold will be ignored and always off
                default: True
        usebg: if false, background color will always be 'default'
                default: True
        colors: number of colors to use (16, 88, or 256)
                default: 256"""
        self.usebold = options.get('usebold', True)
        self.usebg = options.get('usebg', True)
        self.style_attrs = {}
        Formatter.__init__(self, **options)

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, newstyle):
        self._style = newstyle
        self._setup_styles()

    @staticmethod
    def _distance(col1, col2):
        r1, g1, b1 = col1
        r2, g2, b2 = col2

        rd = r1 - r2
        gd = g1 - g2
        bd = b1 - b2

        return rd*rd + gd*gd + bd*bd

    @classmethod
    def findclosest(cls, colstr, colors=256):
        """Takes a hex string and finds the nearest color to it.

        Returns a string urwid will recognize."""

        rgb = int(colstr, 16)
        r = (rgb >> 16) & 0xff
        g = (rgb >> 8) & 0xff
        b = rgb & 0xff

        dist = 257 * 257 * 3
        bestcol = urwid.AttrSpec('h0', 'default')

        for i in range(colors):
            curcol = urwid.AttrSpec('h%d' % i, 'default', colors=colors)
            currgb = curcol.get_rgb_values()[:3]
            curdist = cls._distance((r, g, b), currgb)
            if curdist < dist:
                dist = curdist
                bestcol = curcol

        return bestcol.foreground

    def findclosestattr(self, fgcolstr=None, bgcolstr=None, othersettings='', colors=256):
        """Takes two hex colstring (e.g. 'ff00dd') and returns the
        nearest urwid style."""
        fg = bg = 'default'
        if fgcolstr:
            fg = self.findclosest(fgcolstr, colors)
        if bgcolstr:
            bg = self.findclosest(bgcolstr, colors)
        if othersettings:
            fg = fg + ',' + othersettings
        return urwid.AttrSpec(fg, bg, colors)

    def _setup_styles(self, colors=256):
        """Fills self.style_attrs with urwid.AttrSpec attributes
        corresponding to the closest equivalents to the given style."""
        for ttype, ndef in self.style:
            fgcolstr = bgcolstr = None
            othersettings = ''
            if ndef['color']:
                fgcolstr = ndef['color']
            if self.usebg and ndef['bgcolor']:
                bgcolstr = ndef['bgcolor']
            if self.usebold and ndef['bold']:
                othersettings = 'bold'
            self.style_attrs[str(ttype)] = self.findclosestattr(
                fgcolstr, bgcolstr, othersettings, colors)

    def formatgenerator(self, tokensource):
        """Takes a token source, and generates
        (tokenstring, urwid.AttrSpec) pairs"""
        for (ttype, tstring) in tokensource:
            parts = str(ttype).split(".")
            while str(ttype) not in self.style_attrs:
                parts = parts[:-1]
                ttype = ".".join(parts)

            attr = self.style_attrs[str(ttype)]
            yield attr, tstring

    def format(self, tokensource, outfile):
        for (attr, tstring) in self.formatgenerator(tokensource):
            outfile.write(attr, tstring)



"""
This module contains code for ClickableText
"""


import urwid
from urwid.util import is_mouse_press


class LinkIndicatorSpec(urwid.AttrSpec):
    """Used to track a link within an urwid.Text instance
    """

    def __init__(self, link_label, link_target, orig_spec):
        """Create a new LinkIndicator spec from an existing urwid.AttrSpec

        :param str link_label: The label for the link
        :param str link_target: The target url for the link
        """
        self.link_label = link_label
        self.link_target = link_target

        urwid.AttrSpec.__init__(
            self, orig_spec.foreground, orig_spec.background)


class ClickableText(urwid.Text):
    """Allows clickable/changing text to be part of the Text() contents
    """

    signals = ["click", "change"]

    def mouse_event(self, size, event, button, x, y, focus):
        """Handle mouse events!
        """
        if button != 1 or not is_mouse_press(event):
            return False

        total_offset = (y * size[0]) + x

        text, chunk_stylings = self.get_text()
        curr_offset = 0

        found_style = None
        found_text = None
        found_idx = 0
        found_length = 0

        for idx, info in enumerate(chunk_stylings):
            style, length = info
            if curr_offset < total_offset <= curr_offset + length:
                found_text = text[curr_offset:curr_offset + length]
                found_style = style
                found_idx = idx
                found_length = length
                break
            curr_offset += length

        if found_style is None or not isinstance(found_style, LinkIndicatorSpec):
            self._emit('click')
            return True

        # it's a link, so change the text and update the RLE!
        if found_text == found_style.link_label:
            new_text = found_style.link_target
        else:
            new_text = found_style.link_label
        text = text[:curr_offset] + new_text + text[curr_offset+found_length:]
        new_rle = len(new_text)

        chunk_stylings[found_idx] = (found_style, new_rle)

        self._text = text
        self._attrib = chunk_stylings
        self._invalidate()

        self._emit("change")

        return True



"""
Test the file loader built-in extension
"""


import pytest

import lookatme.config
import lookatme.contrib.file_loader
import lookatme.render.pygments
from tests.utils import assert_render, render_markdown

TEST_STYLE = {
    "style": "monokai",
    "headings": {
        "default": {
            "fg": "bold",
            "bg": "",
            "prefix": "|",
            "suffix": "|",
        },
    },
}


@pytest.fixture(autouse=True)
def file_loader_setup(tmpdir, mocker):
    mocker.patch.object(lookatme.config, "LOG")
    mocker.patch("lookatme.config.SLIDE_SOURCE_DIR", new=str(tmpdir))
    mocker.patch("lookatme.contrib.CONTRIB_MODULES", new=[
        lookatme.contrib.file_loader
    ])
    mocker.patch("lookatme.config.STYLE", new=TEST_STYLE)


def test_file_loader(tmpdir, mocker):
    """Test the built-in file loader extension
    """
    tmppath = tmpdir.join("test.py")
    tmppath.write("print('hello')")

    rendered = render_markdown(f"""
```file
path: {tmppath}
relative: false
```
    """)

    stripped_rows = [
        b'',
        b"print('hello')",
        b'',
        b'',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_file_loader_with_transform(tmpdir, mocker):
    """Test the built-in file loader extension
    """
    tmppath = tmpdir.join("test.py")
    tmppath.write("""
Hello
Apples2
there
Apples3
there
Apples1
""")

    rendered = render_markdown(f"""
```file
path: {tmppath}
relative: false
transform: "grep -i apples | sort"
```
    """)

    stripped_rows = [
        b'',
        b"Apples1",
        b'Apples2',
        b'Apples3',
        b'',
        b'',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_file_loader_relative(tmpdir, mocker):
    """Test the built-in file loader extension
    """
    tmppath = tmpdir.join("test.py")
    tmppath.write("print('hello')")

    rendered = render_markdown("""
```file
path: test.py
relative: true
```
    """)

    stripped_rows = [
        b'',
        b"print('hello')",
        b'',
        b'',
        b'',
    ]
    assert_render(stripped_rows, rendered)


def test_file_loader_not_found(mocker):
    """Test the built-in file loader extension
    """
    rendered = render_markdown("""
```file
path: does_not_exist.py
```
    """)

    stripped_rows = [
        b'',
        b"File not found",
        b'',
        b'',
        b'',
    ]
    assert_render(stripped_rows, rendered)



"""
Setup for lookatme
"""


from setuptools import setup, find_packages
import os


req_path = os.path.join(os.path.dirname(__file__), "requirements.txt")
with open(req_path, "r") as f:
    required = f.read().splitlines()


readme_path = os.path.join(os.path.dirname(__file__), "README.md")
with open(readme_path, "r") as f:
    readme = f.read()


setup(
    name="lookatme",
    version="2.4.1",
    description="An interactive, command-line presentation tool",
    author="James Johnson",
    author_email="d0c.s4vage@gmail.com",
    url="https://github.com/d0c-s4vage/lookatme",
    long_description=readme,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    packages=find_packages(exclude=["docs", ".gitignore", "README.md", "tests"]),
    install_requires=required,
    classifiers=[
        "Environment :: Console",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Multimedia :: Graphics :: Presentation",
        "Topic :: Software Development :: Documentation",
    ],
    entry_points={
        "console_scripts": [
            "lookatme = lookatme.__main__:main",
            "lam = lookatme.__main__:main",
            "witnessme = lookatme.__main__:main",
        ]
    },
)



"""
Slide info holder
"""


class Slide(object):
    """This class defines a single slide. It operates on mistune's lexed
    tokens from the input markdown
    """

    def __init__(self, tokens, number=0):
        """Create a new Slide instance with the provided tokens

        :param list tokens: A list of mistune tokens
        :param int number: The slide number
        """
        self.tokens = tokens
        self.number = number



"""
Defines Presentation specific objects
"""


import os
import threading
import time

import lookatme.ascii_art
import lookatme.config
import lookatme.contrib
import lookatme.prompt
import lookatme.themes
import lookatme.tui
from lookatme.parser import Parser
from lookatme.utils import dict_deep_update


class Presentation(object):
    """Defines a presentation
    """

    def __init__(self, input_stream, theme, style_override=None, live_reload=False,
                 single_slide=False, preload_extensions=None, safe=False,
                 no_ext_warn=False, ignore_ext_failure=False):
        """Creates a new Presentation

        :param stream input_stream: An input stream from which to read the
            slide data
        """
        self.preload_extensions = preload_extensions or []
        self.input_filename = None
        if hasattr(input_stream, "name"):
            lookatme.config.SLIDE_SOURCE_DIR = os.path.dirname(
                input_stream.name)
            self.input_filename = input_stream.name

        self.style_override = style_override
        self.live_reload = live_reload
        self.tui = None
        self.single_slide = single_slide
        self.safe = safe
        self.no_ext_warn = no_ext_warn
        self.ignore_ext_failure = ignore_ext_failure
        self.initial_load_complete = False

        self.theme_mod = __import__(
            "lookatme.themes." + theme, fromlist=[theme])

        if self.live_reload:
            self.reload_thread = threading.Thread(target=self.reload_watcher)
            self.reload_thread.daemon = True
            self.reload_thread.start()

        self.reload(data=input_stream.read())
        self.initial_load_complete = True

    def reload_watcher(self):
        """Watch for changes to the input filename, automatically reloading
        when the modified time has changed.
        """
        if self.input_filename is None:
            return

        last_mod_time = os.path.getmtime(self.input_filename)
        while True:
            try:
                curr_mod_time = os.path.getmtime(self.input_filename)
                if curr_mod_time != last_mod_time:
                    self.get_tui().reload()
                    self.get_tui().loop.draw_screen()
                    last_mod_time = curr_mod_time
            except Exception:
                pass
            finally:
                time.sleep(0.25)

    def reload(self, data=None):
        """Reload this presentation

        :param str data: The data to render for this slide deck (optional)
        """
        if data is None:
            with open(str(self.input_filename), "r") as f:
                data = f.read()

        parser = Parser(single_slide=self.single_slide)
        self.meta, self.slides = parser.parse(data)

        # only load extensions once! Live editing does not support
        # auto-extension reloading
        if not self.initial_load_complete:
            safe_exts = set(self.preload_extensions)
            new_exts = set()
            # only load if running with safe=False
            if not self.safe:
                source_exts = set(self.meta.get("extensions", []))
                new_exts = source_exts - safe_exts
                self.warn_exts(new_exts)

            all_exts = safe_exts | new_exts

            lookatme.contrib.load_contribs(
                all_exts,
                safe_exts,
                self.ignore_ext_failure,
            )

        # style override order:
        # 1. theme settings
        self.styles = lookatme.themes.ensure_defaults(self.theme_mod)
        # 2. inline styles from the presentation
        dict_deep_update(self.styles, self.meta.get("styles", {}))
        # 3. CLI style overrides
        if self.style_override is not None:
            self.styles["style"] = self.style_override  # type: ignore

        lookatme.config.STYLE = self.styles
        self.initial_load_complete = True

    def warn_exts(self, exts):
        """Warn about source-provided extensions that are to-be-loaded
        """
        if len(exts) == 0 or self.no_ext_warn:
            return

        warning = lookatme.ascii_art.WARNING
        print("\n".join(["    " + x for x in warning.split("\n")]))

        print("New extensions required by {!r} are about to be loaded:\n".format(
            self.input_filename
        ))
        for ext in exts:
            print("  - {!r}".format("lookatme.contrib." + ext))
        print("")

        if not lookatme.prompt.yes("Are you ok with attempting to load them?"):
            print("Bailing due to unacceptance of source-required extensions")
            exit(1)

    def run(self, start_slide=0):
        """Run the presentation!
        """
        self.tui = lookatme.tui.create_tui(self, start_slide=start_slide)
        self.tui.run()

    def get_tui(self) -> lookatme.tui.MarkdownTui:
        if self.tui is None:
            raise ValueError(
                "Tui has not been set, has the presentation been run yet?")
        return self.tui



"""
This module defines the text user interface (TUI) for lookatme
"""


import copy
import threading
import time
from collections import defaultdict
from queue import Queue

import urwid

import lookatme.config
import lookatme.config as config
import lookatme.contrib
import lookatme.render.markdown_block as lam_md
from lookatme.contrib import contrib_first
from lookatme.utils import pile_or_listbox_add, spec_from_style


def text(style, data, align="left"):
    if isinstance(style, dict):
        style = spec_from_style(style)
    return urwid.Text((style, data), align=align)


@contrib_first
def root_urwid_widget(to_wrap):
    """This function is overridable by contrib extensions that need to specify
    the root urwid widget.

    The return value *must* return either the ``to_wrap`` widget itself, or
    another widget that wraps the provided ``to_wrap`` widget.
    """
    return to_wrap


class SlideRenderer(threading.Thread):
    daemon = True

    def __init__(self, loop):
        threading.Thread.__init__(self)
        self.events = defaultdict(threading.Event)
        self.keep_running = threading.Event()
        self.queue = Queue()
        self.loop = loop
        self.cache = {}
        self._log = lookatme.config.get_log().getChild("RENDER")

    def flush_cache(self):
        """Clea everything out of the queue and the cache.
        """
        # clear all pending items
        with self.queue.mutex:
            self.queue.queue.clear()
        self.cache.clear()

    def queue_render(self, slide):
        """Queue up a slide to be rendered.
        """
        self.events[slide.number].clear()
        self.queue.put(slide)

    def render_slide(self, slide, force=False):
        """Render a slide, blocking until the slide completes. If ``force`` is
        True, rerender the slide even if it is in the cache.
        """
        if force or slide.number not in self.cache:
            self.events[slide.number].clear()
            self.queue.put(slide)
            self.events[slide.number].wait()

        res = self.cache[slide.number]
        if isinstance(res, Exception):
            raise res
        return res

    def _propagate_meta(self, item1, item2):
        """Copy the metadata from item1 to item2
        """
        meta = getattr(item1, "meta", {})
        existing_meta = getattr(item2, "meta", {})
        new_meta = copy.deepcopy(meta)
        new_meta.update(existing_meta)
        setattr(item2, "meta", new_meta)

    def stop(self):
        self.keep_running.clear()

    def run(self):
        """Run the main render thread
        """
        self.keep_running.set()
        while self.keep_running.is_set():
            to_render = self.queue.get()
            slide_num = to_render.number

            try:
                res = self.do_render(to_render, slide_num)
                self.cache[slide_num] = res
            except Exception as e:
                self.cache[slide_num] = e
            finally:
                self.events[slide_num].set()

    def do_render(self, to_render, slide_num):
        """Perform the actual rendering of a slide. This is done by:

          * parsing the slide into tokens (should have occurred already)
          * iterating through each parsed markdown token
          * calling the appropriately-named render function for the ``token["type"]``
            in :py:mod:`lookatme.render.markdown_block`

        Each render function must have the signature:

        .. code-block:: python

            def render_XXX(token, body, stack, loop):
                pass

        The arguments to the render function are described below:

          * ``token`` - the lexed markdown token - a dictionary
          * ``body`` - the current ``urwid.Pile()`` that return values will be
            added to (same as ``stack[-1]``)
          * ``stack`` - The stack of ``urwid.Pile()`` used during rendering.
            E.g., when rendering nested lists, each nested list will push a new
            ``urwid.Pile()`` to the stack, each wrapped with its own additional
            indentation.
          * ``loop`` - the ``urwid.MainLoop`` instance being used by lookatme.
            This won't usually be used, but is available if needed.

        Main render functions (those defined in markdown_block.py) may have
        three types of return values:

          * ``None`` - nothing is added to ``stack[-1]``. Perhaps the render
            function only needed to add additional indentation by pushing a new
            ``urwid.Pile()`` to the stack.
          * ``list(urwid.Widget)`` - A list of widgets to render. These will
            automatically be added to the Pile at ``stack[-1]``
          * ``urwid.Widget`` - A single widget to render. Will be added to
            ``stack[-1]`` automatically.
        """
        self._log.debug(f"Rendering slide {slide_num}")
        start = time.time()

        # initial processing loop - results are discarded, but render functions
        # may add extra metadata to the token itself. For example, list rendering
        # uses this to determine the max indent size for each level.
        tokens = to_render.tokens
        self._render_tokens(tokens)
        res = self._render_tokens(tokens)

        total = time.time() - start
        self._log.debug(f"Rendered slide {slide_num} in {total}")

        return res

    def _render_tokens(self, tokens):
        tmp_listbox = urwid.ListBox([])
        stack = [tmp_listbox]
        for token in tokens:
            self._log.debug(f"{'  '*len(stack)}Rendering token {token}")

            last_stack = stack[-1]
            last_stack_len = len(stack)

            render_token = getattr(lam_md, f"render_{token['type']}")
            res = render_token(token, stack[-1], stack, self.loop)
            if len(stack) > last_stack_len:
                self._propagate_meta(last_stack, stack[-1])
            if res is None:
                continue
            pile_or_listbox_add(last_stack, res)

        return tmp_listbox.body


class MarkdownTui(urwid.Frame):
    def __init__(self, pres, start_idx=0):
        """
        """
        self.slide_body = urwid.ListBox(
            urwid.SimpleFocusListWalker([urwid.Text("test")]))
        self.slide_title = text("", "", "center")
        self.top_spacing = urwid.Filler(self.slide_title, top=0, bottom=0)
        self.top_spacing_box = urwid.BoxAdapter(self.top_spacing, 1)

        self.creation = text("", "")
        self.slide_num = text("", " test ", "right")
        self.slide_footer = urwid.Columns([self.creation, self.slide_num])
        self.bottom_spacing = urwid.Filler(self.slide_footer, top=0, bottom=0)
        self.bottom_spacing_box = urwid.BoxAdapter(self.bottom_spacing, 1)

        self._log = lookatme.config.get_log()

        urwid.set_encoding('utf8')
        screen = urwid.raw_display.Screen()
        screen.set_terminal_properties(colors=256)

        self.root_margins = urwid.Padding(self, left=2, right=2)
        self.root_paddings = urwid.Padding(self.slide_body, left=10, right=10)

        root_widget = root_urwid_widget(self.root_margins)
        self.loop = urwid.MainLoop(
            root_widget,
            screen=screen,
        )

        # used to track slides that are being rendered
        self.slide_renderer = SlideRenderer(self.loop)
        self.slide_renderer.start()

        self.pres = pres
        self.prep_pres(self.pres, start_idx)

        urwid.Frame.__init__(
            self,
            self.root_paddings,
            self.top_spacing_box,
            self.bottom_spacing_box,
        )

    def prep_pres(self, pres, start_idx=0):
        """Prepare the presentation for displaying/use
        """
        self.curr_slide = self.pres.slides[start_idx]
        self.update()

        # now queue up the rest of the slides while we're at it so they'll be
        # ready when we need them
        for slide in filter(lambda x: x.number != start_idx, self.pres.slides):
            self.slide_renderer.queue_render(slide)

    def update_slide_num(self):
        """Update the slide number
        """
        slide_text = "slide {} / {}".format(
            self.curr_slide.number + 1,
            len(self.pres.slides),
        )
        spec = spec_from_style(config.get_style()["slides"])
        self.slide_num.set_text([(spec, slide_text)])

    def update_title(self):
        """Update the title
        """
        title = self.pres.meta.get("title", "")
        spec = spec_from_style(config.get_style()["title"])
        self.slide_title.set_text([(spec, f" {title} ")])

    def update_creation(self):
        """Update the author and date
        """
        author = self.pres.meta.get('author', '')
        author_spec = spec_from_style(config.get_style()["author"])

        date = self.pres.meta.get('date', '')
        date_spec = spec_from_style(config.get_style()["date"])

        self.creation.set_text([
            (author_spec, f"  {author} "),
            (date_spec, f" {date} "),
        ])

    def update_body(self):
        """Render the provided slide body
        """
        rendered = self.slide_renderer.render_slide(self.curr_slide)
        self.slide_body.body = rendered

    def update_slide_settings(self):
        """Update the slide margins and paddings
        """
        margin = config.get_style()["margin"]
        padding = config.get_style()["padding"]

        self.root_margins.left = margin["left"]
        self.root_margins.right = margin["right"]

        self.root_paddings.left = padding["left"]
        self.root_paddings.right = padding["right"]

        self.top_spacing.top = margin["top"]
        self.top_spacing.bottom = padding["top"]
        self.top_spacing_box.height = margin["top"] + 1 + padding["top"]

        self.bottom_spacing.top = padding["bottom"]
        self.bottom_spacing.bottom = margin["bottom"]
        self.bottom_spacing_box.height = margin["bottom"] + \
            1 + padding["bottom"]

    def update(self):
        """
        """
        self.update_slide_settings()
        self.update_slide_num()
        self.update_title()
        self.update_creation()
        self.update_body()

    def reload(self):
        """Reload the input, keeping the current slide in focus
        """
        curr_slide_idx = self.curr_slide.number
        self.slide_renderer.flush_cache()
        self.pres.reload()
        self.prep_pres(self.pres, curr_slide_idx)
        self.update()

    def keypress(self, size, key):
        """Handle keypress events
        """
        self._log.debug(f"KEY: {key}")
        key = self._get_key(size, key)
        if key is None:
            return

        slide_direction = 0
        if key in ["left", "backspace", "delete", "h", "k"]:
            slide_direction = -1
        elif key in ["right", " ", "j", "l"]:
            slide_direction = 1
        elif key in ["q", "Q"]:
            lookatme.contrib.shutdown_contribs()
            raise urwid.ExitMainLoop()
        elif key == "r":
            self.reload()

        if slide_direction == 0:
            return

        new_slide_num = self.curr_slide.number + slide_direction
        if new_slide_num < 0:
            new_slide_num = 0
        elif new_slide_num >= len(self.pres.slides):
            new_slide_num = len(self.pres.slides) - 1

        if new_slide_num == self.curr_slide.number:
            return

        self.curr_slide = self.pres.slides[new_slide_num]
        self.update()

    def _get_key(self, size, key):
        """Resolve the key that was pressed.
        """
        try:
            key = urwid.Frame.keypress(self, size, key)
            if key is None:
                return
        except Exception:
            pass
        return key

    def run(self):
        self.loop.run()


def create_tui(pres, start_slide=0):
    """Run the provided presentation

    :param int start_slide: 0-based slide index
    """
    tui = MarkdownTui(pres, start_slide)
    return tui



"""
Misc ASCII art
"""

WARNING = r"""
               _mBma
              sQf "QL
             jW(   -$g.
            jW'     -$m,
          .y@'  _aa.  4m,
         .mD`  ]QQWQ.  4Q,
        _mP`   ]QQQQ    ?Q/
       _QF     )WQQ@     ?Qc
      <QF       QQQF      )Qa
     jW(        QQQf       "QL
    jW'         ]H8'        -Q6.
  .y@'          _as.         -$m.
 .m@`          ]QQWQ.         -4m,
_mP`           -?$8!            4Q,
mE                               $m
?$gyygggggggggwywgyygggggggygggggD(
"""



"""
Defines all schemas used in lookatme
"""


import datetime
from typing import Dict

import pygments.styles
import yaml
from marshmallow import INCLUDE, RAISE, Schema, fields, validate


class NoDatesSafeLoader(yaml.SafeLoader):
    @classmethod
    def remove_implicit_resolver(cls, tag_to_remove):
        """
        Remove implicit resolvers for a particular tag

        Takes care not to modify resolvers in super classes.

        We want to load datetimes as strings, not dates, because we
        go on to serialise as json which doesn't have the advanced types
        of yaml, and leads to incompatibilities down the track.
        """
        if 'yaml_implicit_resolvers' not in cls.__dict__:
            cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()

        for first_letter, mappings in cls.yaml_implicit_resolvers.items():
            cls.yaml_implicit_resolvers[first_letter] = [
                (tag, regexp) for tag, regexp in mappings if tag != tag_to_remove
            ]


NoDatesSafeLoader.remove_implicit_resolver('tag:yaml.org,2002:timestamp')


class YamlRender:
    @staticmethod
    def loads(data): return yaml.load(data, Loader=NoDatesSafeLoader)
    @staticmethod
    def dumps(data): return yaml.safe_dump(data, allow_unicode=True)


class BulletsSchema(Schema):
    default = fields.Str(dump_default="•")

    class Meta:
        include = {
            "1": fields.Str(dump_default="•"),
            "2": fields.Str(dump_default="⁃"),
            "3": fields.Str(dump_default="◦"),
            "4": fields.Str(dump_default="•"),
            "5": fields.Str(dump_default="⁃"),
            "6": fields.Str(dump_default="◦"),
            "7": fields.Str(dump_default="•"),
            "8": fields.Str(dump_default="⁃"),
            "9": fields.Str(dump_default="◦"),
            "10": fields.Str(dump_default="•"),
        }


_NUMBERING_VALIDATION = validate.OneOf(["numeric", "alpha", "roman"])


class NumberingSchema(Schema):
    default = fields.Str(dump_default="numeric",
                         validate=_NUMBERING_VALIDATION)

    class Meta:
        include = {
            "1": fields.Str(dump_default="numeric", validate=_NUMBERING_VALIDATION),
            "2": fields.Str(dump_default="alpha", validate=_NUMBERING_VALIDATION),
            "3": fields.Str(dump_default="roman", validate=_NUMBERING_VALIDATION),
            "4": fields.Str(dump_default="numeric", validate=_NUMBERING_VALIDATION),
            "5": fields.Str(dump_default="alpha", validate=_NUMBERING_VALIDATION),
            "6": fields.Str(dump_default="roman", validate=_NUMBERING_VALIDATION),
            "7": fields.Str(dump_default="numeric", validate=_NUMBERING_VALIDATION),
            "8": fields.Str(dump_default="alpha", validate=_NUMBERING_VALIDATION),
            "9": fields.Str(dump_default="roman", validate=_NUMBERING_VALIDATION),
            "10": fields.Str(dump_default="numeric", validate=_NUMBERING_VALIDATION),
        }


class StyleFieldSchema(Schema):
    fg = fields.Str(dump_default="")
    bg = fields.Str(dump_default="")


class SpacingSchema(Schema):
    top = fields.Int(dump_default=0)
    bottom = fields.Int(dump_default=0)
    left = fields.Int(dump_default=0)
    right = fields.Int(dump_default=0)


class HeadingStyleSchema(Schema):
    prefix = fields.Str()
    suffix = fields.Str()
    fg = fields.Str(dump_default="")
    bg = fields.Str(dump_default="")


class HruleSchema(Schema):
    char = fields.Str(dump_default="─")
    style = fields.Nested(StyleFieldSchema, dump_default=StyleFieldSchema().dump({
        "fg": "#777",
        "bg": "default",
    }))


class BlockQuoteSchema(Schema):
    side = fields.Str(dump_default="╎")
    top_corner = fields.Str(dump_default="┌")
    bottom_corner = fields.Str(dump_default="└")
    style = fields.Nested(StyleFieldSchema, dump_default=StyleFieldSchema().dump({
        "fg": "italics,#aaa",
        "bg": "default",
    }))


class HeadingsSchema(Schema):
    default = fields.Nested(HeadingStyleSchema, dump_default={
        "fg": "#346,bold",
        "bg": "default",
        "prefix": "░░░░░ ",
        "suffix": "",
    })

    class Meta:
        include = {
            "1": fields.Nested(HeadingStyleSchema, dump_default={
                "fg": "#9fc,bold",
                "bg": "default",
                "prefix": "██ ",
                "suffix": "",
            }),
            "2": fields.Nested(HeadingStyleSchema, dump_default={
                "fg": "#1cc,bold",
                "bg": "default",
                "prefix": "▓▓▓ ",
                "suffix": "",
            }),
            "3": fields.Nested(HeadingStyleSchema, dump_default={
                "fg": "#29c,bold",
                "bg": "default",
                "prefix": "▒▒▒▒ ",
                "suffix": "",
            }),
            "4": fields.Nested(HeadingStyleSchema, dump_default={
                "fg": "#559,bold",
                "bg": "default",
                "prefix": "░░░░░ ",
                "suffix": "",
            }),
            "5": fields.Nested(HeadingStyleSchema),
            "6": fields.Nested(HeadingStyleSchema),
        }


class TableSchema(Schema):
    header_divider = fields.Str(dump_default="─")
    column_spacing = fields.Int(dump_default=3)


class StyleSchema(Schema):
    """Styles schema for themes and style overrides within presentations
    """
    class Meta:
        render_module = YamlRender
        unknown = RAISE

    style = fields.Str(
        dump_default="monokai",
        validate=validate.OneOf(list(pygments.styles.get_all_styles())),
    )

    title = fields.Nested(StyleFieldSchema, dump_default={
        "fg": "#f30,bold,italics",
        "bg": "default",
    })
    author = fields.Nested(StyleFieldSchema, dump_default={
        "fg": "#f30",
        "bg": "default",
    })
    date = fields.Nested(StyleFieldSchema, dump_default={
        "fg": "#777",
        "bg": "default",
    })
    slides = fields.Nested(StyleFieldSchema, dump_default={
        "fg": "#f30",
        "bg": "default",
    })
    margin = fields.Nested(SpacingSchema, dump_default={
        "top": 0,
        "bottom": 0,
        "left": 2,
        "right": 2,
    })
    padding = fields.Nested(SpacingSchema, dump_default={
        "top": 0,
        "bottom": 0,
        "left": 10,
        "right": 10,
    })

    headings = fields.Nested(
        HeadingsSchema, dump_default=HeadingsSchema().dump(None))
    bullets = fields.Nested(
        BulletsSchema, dump_default=BulletsSchema().dump(None))
    numbering = fields.Nested(
        NumberingSchema, dump_default=NumberingSchema().dump(None))
    table = fields.Nested(TableSchema, dump_default=TableSchema().dump(None))
    quote = fields.Nested(
        BlockQuoteSchema, dump_default=BlockQuoteSchema().dump(None))
    hrule = fields.Nested(HruleSchema, dump_default=HruleSchema().dump(None))
    link = fields.Nested(StyleFieldSchema, dump_default={
        "fg": "#33c,underline",
        "bg": "default",
    })


class MetaSchema(Schema):
    """The schema for presentation metadata
    """
    class Meta:
        render_module = YamlRender
        unknown = INCLUDE

    title = fields.Str(dump_default="", load_default="")
    date = fields.Str(
        dump_default=datetime.datetime.now().strftime("%Y-%m-%d"),
        load_default=datetime.datetime.now().strftime("%Y-%m-%d"),
    )
    author = fields.Str(dump_default="", load_default="")
    styles = fields.Nested(
        StyleSchema,
        dump_default=StyleSchema().dump(None),
        load_default=StyleSchema().dump(None),
    )
    extensions = fields.List(fields.Str(), dump_default=[], load_default=[])

    def _ensure_top_level_defaults(self, res: Dict) -> Dict:
        res.setdefault("title", "")
        res.setdefault("author", "")
        res.setdefault("date", datetime.datetime.now().strftime("%Y-%m-%d"))
        res.setdefault("extensions", [])

        return res

    def loads_partial_styles(self, *args, **kwargs):
        kwargs["partial"] = True
        res = super(self.__class__, self).loads(*args, **kwargs)
        if res is None:
            raise ValueError("Could not loads")

        res = self._ensure_top_level_defaults(res)
        return res

    def loads(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).loads(*args, **kwargs)
        if res is None:
            raise ValueError("Could not loads")

        return res

    def load_partial_styles(self, *args, **kwargs):
        kwargs["partial"] = True
        res = super(self.__class__, self).load(*args, **kwargs)
        if res is None:
            raise ValueError("Could not loads")

        res = self._ensure_top_level_defaults(res)
        return res

    def load(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).load(*args, **kwargs)
        if res is None:
            raise ValueError("Could not load")

        return res

    def dump(self, *args, **kwargs) -> Dict:
        res = super(self.__class__, self).dump(*args, **kwargs)
        if res is None:
            raise ValueError("Could not dump")
        return res



"""
Defines render functions that work with mistune's markdown inline lexer render
interface
"""


import functools

import lookatme.config as config
import lookatme.render.pygments as pygments_render
import lookatme.utils as utils
from lookatme.contrib import contrib_first
from lookatme.widgets.clickable_text import LinkIndicatorSpec

options = {}


def expanded_styles(fn):
    @functools.wraps(fn)
    def inner(text):
        styles = dict(fg="", bg="")
        if isinstance(text, str):
            return fn(text, styles)
        elif isinstance(text, list) and isinstance(text[0], str):
            return fn(text[0], styles)
        elif isinstance(text, list) and isinstance(text[0], tuple):
            attr_spec = text[0][0]
            styles = dict(fg=attr_spec.foreground, bg=attr_spec.background)
            text = text[0][1]
            return fn(text, styles)
        else:
            return fn(text, styles)
    return inner


# -------------------------------------------------------------------------


def placeholder():
    """The starting point of the rendering. The final result will be this
    returned list with all inline markdown tokens translated into urwid objects
    """
    return []


def render_no_change(text):
    """Render inline markdown text with no changes
    """
    return [text]


@contrib_first
def inline_html(text):
    """Renders inline html as plaintext

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return render_no_change(text)


@contrib_first
def text(text):
    """Renders plain text (does nothing)

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return render_no_change(text)


@contrib_first
def escape(text):
    """Renders escapes

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return render_no_change(text)


@contrib_first
def autolink(link_uri, is_email=False):
    """Renders a URI as a link

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return link(link_uri, None, link_uri)


@contrib_first
def footnote_ref(key, index):
    """Renders a footnote

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return render_no_change(key)


@contrib_first
def image(link_uri, title, text):
    """Renders an image as a link. This would be a cool extension to render
    referenced images as scaled-down ansii pixel blocks.

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return link(link_uri, title, text)


@contrib_first
def link(link_uri, title, link_text):
    """Renders a link. This function does a few special things to make the
    clickable links happen. All text in lookatme is rendered using the
    :any:`ClickableText` class. The ``ClickableText`` class looks for
    ``urwid.AttrSpec`` instances that are actually ``LinkIndicatorSpec`` instances
    within the Text markup. If an AttrSpec is an instance of ``LinkIndicator``
    spec in the Text markup, ClickableText knows to handle clicks on that
    section of the text as a link.

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    raw_link_text = []
    for x in link_text:
        if isinstance(x, tuple):
            raw_link_text.append(x[1])
        else:
            raw_link_text.append(x)
    raw_link_text = "".join(raw_link_text)

    spec, text = utils.styled_text(
        link_text, utils.spec_from_style(config.get_style()["link"]))
    spec = LinkIndicatorSpec(raw_link_text, link_uri, spec)
    return [(spec, text)]


@expanded_styles
@contrib_first
def double_emphasis(text, old_styles):
    """Renders double emphasis. Handles both ``**word**`` and ``__word__``

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return [utils.styled_text(text, "underline", old_styles)]


@expanded_styles
@contrib_first
def emphasis(text, old_styles):
    """Renders double emphasis. Handles both ``*word*`` and ``_word_``

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return [utils.styled_text(text, "italics", old_styles)]


@expanded_styles
@contrib_first
def codespan(text, old_styles):
    """Renders inline code using the pygments renderer. This function also makes
    use of the coding style:

    .. code-block:: yaml

        style: monokai

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    res = pygments_render.render_text(" " + text + " ", plain=True)
    return res


@contrib_first
def linebreak():
    """Renders a line break

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return ["\n"]


@expanded_styles
@contrib_first
def strikethrough(text, old_styles):
    """Renders strikethrough text (``~~text~~``)

    :returns: list of `urwid Text markup <http://urwid.org/manual/displayattributes.html#text-markup>`_
        tuples.
    """
    return [utils.styled_text(text, "strikethrough", old_styles)]
